#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
General purpose separator block for IDAES models
"""

from enum import Enum
from pandas import DataFrame

from pyomo.environ import (Block, Set)
from pyomo.network import Port
from pyomo.common.config import ConfigBlock, ConfigValue, In, ListOf, Bool

from idaes.core import (
    declare_process_block_class,
    UnitModelBlockData,
    useDefault,
    MaterialBalanceType,
    MomentumBalanceType,
)
from idaes.core.util.config import (
    is_physical_parameter_block,
    is_state_block,
)
from idaes.core.util.exceptions import (
    BurntToast,
    ConfigurationError,
)
from idaes.core.solvers import get_solver
from idaes.core.util.tables import create_stream_table_dataframe
from idaes.core.util.model_statistics import degrees_of_freedom
import idaes.logger as idaeslog
import idaes.core.util.scaling as iscale
from idaes.core.util.units_of_measurement import report_quantity
from idaes.core.initialization import ModularInitializerBase

__author__ = "Team Ahuora"


# Set up logger
_log = idaeslog.getLogger(__name__)


# Enumerate options for balances
class SplittingType(Enum):
    """
    Enum of supported material split types.
    """

    totalFlow = 1
    phaseFlow = 2
    componentFlow = 3
    phaseComponentFlow = 4


class EnergySplittingType(Enum):
    """
    Enum of support energy split types.
    """

    none = 0
    equal_molar_enthalpy = 2


class SimpleSeparatorInitializer(ModularInitializerBase):
    """
    Initializer for Separator blocks.

    """

    def initialization_routine(
        self,
        model: Block,
    ):
        """
        Initialization routine for Separator Blocks.

        This routine starts by initializing the feed and outlet streams using simple rules.

        Args:
            model: model to be initialized

        Returns:
            None

        """
        init_log = idaeslog.getInitLogger(
            model.name, self.get_output_level(), tag="unit"
        )
        solve_log = idaeslog.getSolveLogger(
            model.name, self.get_output_level(), tag="unit"
        )

        # Create solver
        solver = self._get_solver()
        # Initialize mixed state block
        
        mblock = model.mixed_state
        self.get_submodel_initializer(mblock).initialize(mblock)

        res = None
        if degrees_of_freedom(model) != 0:
            with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
                res = solver.solve(model, tee=slc.tee)
                init_log.info(
                    "Initialization Step 1 Complete: {}".format(idaeslog.condition(res))
                )

        for c, s in component_status.items():
            if s:
                c.activate()


        # Initialize outlet StateBlocks
        outlet_list = model._create_outlet_list()

        # Initializing outlet states
        for o in outlet_list:
            # Get corresponding outlet StateBlock
            o_block = getattr(model, o + "_state")

            # Create dict to store fixed status of state variables
            for t in model.flowsheet().time:
                # Calculate values for state variables
                s_vars = o_block[t].define_state_vars()
                for var_name_port, var_obj in s_vars.items():
                    for k in var_obj:
                        # If fixed, use current value
                        # otherwise calculate guess from mixed state and fix
                        if not var_obj[k].fixed:
                            m_var = getattr(mblock[t], var_obj.local_name)
                            if "flow" in var_name_port:
                                # Leave initial value
                                pass
                            else:
                                # Otherwise intensive, equate to mixed stream
                                var_obj[k].set_value(m_var[k].value)

            # Call initialization routine for outlet StateBlock
            self.get_submodel_initializer(o_block).initialize(o_block)

        init_log.info("Initialization Complete.")

        return res


@declare_process_block_class("SimpleSeparator")
class SimpleSeparatorData(UnitModelBlockData):
    """
    This is a simple Splitter block with the IDAES modeling framework. 
    Unlike the generic Separator, this block avoids use of split fractions.

    This model creates a number of StateBlocks to represent the outgoing
    streams, then writes a set of phase-component material balances, an
    overall enthalpy balance (2 options), and a momentum balance (2 options)
    linked to a mixed-state StateBlock. The mixed-state StateBlock can either
    be specified by the user (allowing use as a sub-model), or created by the
    Separator.
    """

    default_initializer = SimpleSeparatorInitializer

    CONFIG = UnitModelBlockData.CONFIG()
    CONFIG.declare(
        "property_package",
        ConfigValue(
            default=useDefault,
            domain=is_physical_parameter_block,
            description="Property package to use for mixer",
            doc="""Property parameter object used to define property
calculations,
**default** - useDefault.
**Valid values:** {
**useDefault** - use default package from parent model or flowsheet,
**PropertyParameterObject** - a PropertyParameterBlock object.}""",
        ),
    )
    CONFIG.declare(
        "property_package_args",
        ConfigBlock(
            implicit=True,
            description="Arguments to use for constructing property packages",
            doc="""A ConfigBlock with arguments to be passed to a property
block(s) and used when constructing these,
**default** - None.
**Valid values:** {
see property package for documentation.}""",
        ),
    )
    CONFIG.declare(
        "outlet_list",
        ConfigValue(
            domain=ListOf(str),
            description="List of outlet names",
            doc="""A list containing names of outlets,
**default** - None.
**Valid values:** {
**None** - use num_outlets argument,
**list** - a list of names to use for outlets.}""",
        ),
    )
    CONFIG.declare(
        "num_outlets",
        ConfigValue(
            domain=int,
            description="Number of outlets to unit",
            doc="""Argument indicating number (int) of outlets to construct,
not used if outlet_list arg is provided,
**default** - None.
**Valid values:** {
**None** - use outlet_list arg instead, or default to 2 if neither argument
provided,
**int** - number of outlets to create (will be named with sequential integers
from 1 to num_outlets).}""",
        ),
    )
    CONFIG.declare(
        "material_balance_type",
        ConfigValue(
            default=MaterialBalanceType.useDefault,
            domain=In(MaterialBalanceType),
            description="Material balance construction flag",
            doc="""Indicates what type of mass balance should be constructed,
**default** - MaterialBalanceType.useDefault.
**Valid values:** {
**MaterialBalanceType.useDefault - refer to property package for default
balance type
**MaterialBalanceType.none** - exclude material balances,
**MaterialBalanceType.componentPhase** - use phase component balances,
**MaterialBalanceType.componentTotal** - use total component balances,
**MaterialBalanceType.elementTotal** - use total element balances,
**MaterialBalanceType.total** - use total material balance.}""",
        ),
    )
    CONFIG.declare(
        "momentum_balance_type",
        ConfigValue(
            default=MomentumBalanceType.pressureTotal,
            domain=In(MomentumBalanceType),
            description="Momentum balance construction flag",
            doc="""Indicates what type of momentum balance should be constructed,
    **default** - MomentumBalanceType.pressureTotal.
    **Valid values:** {
    **MomentumBalanceType.none** - exclude momentum balances,
    **MomentumBalanceType.pressureTotal** - pressure in all outlets is equal,
    **MomentumBalanceType.pressurePhase** - not yet supported,
    **MomentumBalanceType.momentumTotal** - not yet supported,
    **MomentumBalanceType.momentumPhase** - not yet supported.}""",
        ),
    )
    CONFIG.declare(
        "has_phase_equilibrium",
        ConfigValue(
            default=False,
            domain=Bool,
            description="Calculate phase equilibrium in mixed stream",
            doc="""Argument indicating whether phase equilibrium should be
calculated for the resulting mixed stream,
**default** - False.
**Valid values:** {
**True** - calculate phase equilibrium in mixed stream,
**False** - do not calculate equilibrium in mixed stream.}""",
        ),
    )

    def build(self):
        """
        General build method for SeparatorData. This method calls a number
        of sub-methods which automate the construction of expected attributes
        of unit models.

        Inheriting models should call `super().build`.

        Args:
            None

        Returns:
            None
        """
        # Call super.build()
        super(SimpleSeparatorData, self).build()

        # Call setup methods from ControlVolumeBlockData
        self._get_property_package()
        self._get_indexing_sets()

        # Create list of inlet names
        outlet_list = self._create_outlet_list()

        mixed_block = self._add_mixed_state_block()

        # Add inlet port
        self._add_inlet_port_objects(mixed_block)

        # Build StateBlocks for outlet
        outlet_blocks = self._add_outlet_state_blocks(outlet_list)
        self.outlet_idx = Set(initialize=outlet_list)

        # Construct splitting equations
        self._add_material_balance(mixed_block, outlet_blocks)
        self._add_energy_balance(mixed_block, outlet_blocks)
        self._add_momentum_balance(mixed_block, outlet_blocks)

        # Construct outlet port objects
        self._add_outlet_port_objects(outlet_list)

    def _create_outlet_list(self):
        """
        Create list of outlet stream names based on config arguments.

        Returns:
            list of strings
        """
        if self.config.outlet_list is not None and self.config.num_outlets is not None:
            # If both arguments provided and not consistent, raise Exception
            if len(self.config.outlet_list) != self.config.num_outlets:
                raise ConfigurationError(
                    "{} Separator provided with both outlet_list and "
                    "num_outlets arguments, which were not consistent ("
                    "length of outlet_list was not equal to num_outlets). "
                    "Please check your arguments for consistency, and "
                    "note that it is only necessry to provide one of "
                    "these arguments.".format(self.name)
                )
        elif self.config.outlet_list is None and self.config.num_outlets is None:
            # If no arguments provided for outlets, default to num_outlets = 2
            self.config.num_outlets = 2

        # Create a list of names for outlet StateBlocks
        if self.config.outlet_list is not None:
            outlet_list = self.config.outlet_list
        else:
            outlet_list = [
                "outlet_" + str(n) for n in range(1, self.config.num_outlets + 1)
            ]

        return outlet_list

    def _add_outlet_state_blocks(self, outlet_list):
        """
        Construct StateBlocks for all outlet streams.

        Args:
            list of strings to use as StateBlock names

        Returns:
            list of StateBlocks
        """
        # Setup StateBlock argument dict
        tmp_dict = dict(**self.config.property_package_args)
        tmp_dict["has_phase_equilibrium"] = False
        tmp_dict["defined_state"] = False

        # Create empty list to hold StateBlocks for return
        outlet_blocks = []

        # Create an instance of StateBlock for all outlets
        for o in outlet_list:
            o_obj = self.config.property_package.build_state_block(
                self.flowsheet().time, doc="Material properties at outlet", **tmp_dict
            )

            setattr(self, o + "_state", o_obj)

            outlet_blocks.append(getattr(self, o + "_state"))

        return outlet_blocks

    def _add_mixed_state_block(self):
        """
        Constructs StateBlock to represent mixed stream.

        Returns:
            New StateBlock object
        """
        # Setup StateBlock argument dict
        tmp_dict = dict(**self.config.property_package_args)
        tmp_dict["has_phase_equilibrium"] = False
        tmp_dict["defined_state"] = True

        self.mixed_state = self.config.property_package.build_state_block(
            self.flowsheet().time, doc="Material properties of mixed stream", **tmp_dict
        )

        return self.mixed_state

    def _add_inlet_port_objects(self, mixed_block):
        """ Adds inlet Port object."""
        self.add_port(name="inlet", block=mixed_block, doc="Inlet Port")

    def _add_outlet_port_objects(self, outlet_list):
        """Adds outlet Port objects."""
        for p in outlet_list:
            o_state = getattr(self, p + "_state")
            self.add_port(name=p, block=o_state, doc="Outlet Port")

    def _add_material_balance(self, mixed_block, outlet_blocks):
        """Add overall material balance equation."""
        # Get phase component list(s)
        pc_set = mixed_block.phase_component_set
        
        # Write phase-component balances
        @self.Constraint(self.flowsheet().time, doc="Material balance equation")
        def material_balance_equation(b, t):
            return 0 == sum(
                sum(
                    mixed_block[t].get_material_flow_terms(p, j)
                    - 
                    sum(
                        o[t].get_material_flow_terms(p, j)
                        for o in outlet_blocks
                    )                    
                    for j in mixed_block.component_list
                    if (p, j) in pc_set
                )
                for p in mixed_block.phase_list
            )

    def _add_energy_balance(self, mixed_block, outlet_blocks):
        """
        Creates constraints for splitting the energy flows.
        """
        # split basis is equal_molar_enthalpy
        @self.Constraint(
            self.flowsheet().time,
            self.outlet_idx,
            doc="Molar enthalpy equality constraint",
        )
        def molar_enthalpy_equality_eqn(b, t, o):
            o_block = getattr(self, o + "_state")
            return mixed_block[t].enth_mol == o_block[t].enth_mol

    def _add_momentum_balance(self, mixed_block, outlet_blocks):
        """
        Creates constraints for splitting the momentum flows - done by equating
        pressures in outlets.
        """
        if self.config.momentum_balance_type is MomentumBalanceType.pressureTotal:
            @self.Constraint(
                self.flowsheet().time,
                self.outlet_idx,
                doc="Pressure equality constraint",
            )
            def pressure_equality_eqn(b, t, o):
                o_block = getattr(self, o + "_state")
                return mixed_block[t].pressure == o_block[t].pressure

    def model_check(blk):
        """
        This method executes the model_check methods on the associated state
        blocks (if they exist). This method is generally called by a unit model
        as part of the unit's model_check method.

        Args:
            None

        Returns:
            None
        """
        # Try property block model check
        for t in blk.flowsheet().time:
            try:
                blk.mixed_state[t].model_check()
            except AttributeError:
                _log.warning(
                    "{} Separator inlet state block has no "
                    "model check. To correct this, add a "
                    "model_check method to the associated "
                    "StateBlock class.".format(blk.name)
                )

            try:
                outlet_list = blk._create_outlet_list()
                for o in outlet_list:
                    o_block = getattr(blk, o + "_state")
                    o_block[t].model_check()
            except AttributeError:
                _log.warning(
                    "{} Separator outlet state block has no "
                    "model checks. To correct this, add a model_check"
                    " method to the associated StateBlock class.".format(blk.name)
                )

    def initialize_build(
        blk, outlvl=idaeslog.NOTSET, optarg=None, solver=None, hold_state=False
    ):
        """
        Initialization routine for separator

        Keyword Arguments:
            outlvl : sets output level of initialization routine
            optarg : solver options dictionary object (default=None, use
                     default solver options)
            solver : str indicating which solver to use during
                     initialization (default = None, use default solver)
            hold_state : flag indicating whether the initialization routine
                     should unfix any state variables fixed during
                     initialization, **default** - False. **Valid values:**
                     **True** - states variables are not unfixed, and a dict of
                     returned containing flags for which states were fixed
                     during initialization, **False** - state variables are
                     unfixed after initialization by calling the release_state
                     method.

        Returns:
            If hold_states is True, returns a dict containing flags for which
            states were fixed during initialization.
        """
        init_log = idaeslog.getInitLogger(blk.name, outlvl, tag="unit")
        solve_log = idaeslog.getSolveLogger(blk.name, outlvl, tag="unit")

        # Create solver
        opt = get_solver(solver, optarg)

        mblock = blk.mixed_state
        flags = mblock.initialize(
            outlvl=outlvl,
            optarg=optarg,
            solver=solver,
            hold_state=True,
        )

        if degrees_of_freedom(blk) != 0:
            with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
                res = opt.solve(blk, tee=slc.tee)
                init_log.info(
                    "Initialization Step 1 Complete: {}".format(idaeslog.condition(res))
                )

        # Initialize outlet StateBlocks
        outlet_list = blk._create_outlet_list()

        # Premises for initializing outlet states:
        for o in outlet_list:
            # Get corresponding outlet StateBlock
            o_block = getattr(blk, o + "_state")

            # Create dict to store fixed status of state variables
            o_flags = {}
            for t in blk.flowsheet().time:

                # Calculate values for state variables
                s_vars = o_block[t].define_state_vars()
                for v in s_vars:
                    for k in s_vars[v]:
                        # Record whether variable was fixed or not
                        o_flags[t, v, k] = s_vars[v][k].fixed

                        # If fixed, use current value
                        # otherwise calculate guess from mixed state and fix
                        if not s_vars[v][k].fixed:
                            m_var = getattr(mblock[t], s_vars[v].local_name)
                            if "flow" in v:
                                # Leave initial value, but avoid negative flows
                                if s_vars[v][k].value < 1e-4:
                                    s_vars[v][k].set_value(1e-2)
                            else:
                                # Otherwise intensive, equate to mixed stream
                                s_vars[v][k].set_value(m_var[k].value)

            # Call initialization routine for outlet StateBlock
            o_block.initialize(
                outlvl=outlvl,
                optarg=optarg,
                solver=solver,
                hold_state=False,
            )

            # Revert fixed status of variables to what they were before
            for t in blk.flowsheet().time:
                s_vars = o_block[t].define_state_vars()
                for v in s_vars:
                    for k in s_vars[v]:
                        s_vars[v][k].fixed = o_flags[t, v, k]

        init_log.info("Initialization Complete.")
        return flags

    def release_state(blk, flags, outlvl=idaeslog.NOTSET):
        """
        Method to release state variables fixed during initialization.

        Keyword Arguments:
            flags : dict containing information of which state variables
                    were fixed during initialization, and should now be
                    unfixed. This dict is returned by initialize if
                    hold_state = True.
            outlvl : sets output level of logging

        Returns:
            None
        """
        mblock = blk.mixed_state
        mblock.release_state(flags, outlvl=outlvl)

    def calculate_scaling_factors(self):
        mb_type = self.config.material_balance_type
        mixed_state = self.mixed_state
        if mb_type == MaterialBalanceType.useDefault:
            t_ref = self.flowsheet().time.first()
            mb_type = mixed_state[t_ref].default_material_balance_type()
        super().calculate_scaling_factors()

        if hasattr(self, "temperature_equality_eqn"):
            for (t, i), c in self.temperature_equality_eqn.items():
                s = iscale.get_scaling_factor(
                    mixed_state[t].temperature, default=1, warning=True
                )
                iscale.constraint_scaling_transform(c, s)

        if hasattr(self, "pressure_equality_eqn"):
            for (t, i), c in self.pressure_equality_eqn.items():
                s = iscale.get_scaling_factor(
                    mixed_state[t].pressure, default=1, warning=True
                )
                iscale.constraint_scaling_transform(c, s)

        if hasattr(self, "material_splitting_eqn"):
            if mb_type == MaterialBalanceType.componentPhase:
                for (t, _, p, j), c in self.material_splitting_eqn.items():
                    flow_term = mixed_state[t].get_material_flow_terms(p, j)
                    s = iscale.get_scaling_factor(flow_term, default=1)
                    iscale.constraint_scaling_transform(c, s)
            elif mb_type == MaterialBalanceType.componentTotal:
                for (t, _, j), c in self.material_splitting_eqn.items():
                    s = None
                    for p in mixed_state.phase_list:
                        try:
                            ft = mixed_state[t].get_material_flow_terms(p, j)
                        except KeyError:
                            # This component does not exist in this phase
                            continue
                        if s is None:
                            s = iscale.get_scaling_factor(ft, default=1)
                        else:
                            _s = iscale.get_scaling_factor(ft, default=1)
                            s = _s if _s < s else s
                    iscale.constraint_scaling_transform(c, s)
            elif mb_type == MaterialBalanceType.total:
                pc_set = mixed_state.phase_component_set
                for (t, _), c in self.material_splitting_eqn.items():
                    for i, (p, j) in enumerate(pc_set):
                        ft = mixed_state[t].get_material_flow_terms(p, j)
                        if i == 0:
                            s = iscale.get_scaling_factor(ft, default=1)
                        else:
                            _s = iscale.get_scaling_factor(ft, default=1)
                            s = _s if _s < s else s
                    iscale.constraint_scaling_transform(c, s)

    def _get_performance_contents(self, time_point=0):
        if hasattr(self, "split_fraction"):
            var_dict = {}
            for k, v in self.split_fraction.items():
                if k[0] == time_point:
                    var_dict[f"Split Fraction [{str(k[1:])}]"] = v
            return {"vars": var_dict}
        else:
            return None

    def _get_stream_table_contents(self, time_point=0):
        outlet_list = self._create_outlet_list()

        io_dict = {}
        io_dict["Inlet"] = self.mixed_state

        for o in outlet_list:
            io_dict[o] = getattr(self, o + "_state")

        return create_stream_table_dataframe(io_dict, time_point=time_point)



from ..custom_tank import DynamicTank
import pytest
import pyomo.environ as pyo
from pyomo.network import Arc
from idaes.core import FlowsheetBlock, MaterialBalanceType
from idaes.models.unit_models import Heater, Valve
from idaes.models.properties import iapws95
from idaes.core.util.initialization import propagate_state
from idaes.core.util.model_statistics import degrees_of_freedom
# need to pip install ahuora_compounds@git+https://github.com/waikato-ahuora-smart-energy-systems/PropertyPackages.git@v0.0.25
from property_packages.build_package import build_package

def test_dynamic_tank():
    m = pyo.ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=True,time_set=[0,1,2,3],time_units=pyo.units.s)
    m.fs.water = build_package("helmholtz", ["water"],["Liq","Vap"])
    m.fs.tank = DynamicTank(property_package=m.fs.water,    tank_type="vertical_cylindrical_tank", has_holdup=True,
    has_heat_transfer=True, dynamic=True)

    m.discretizer = pyo.TransformationFactory("dae.finite_difference")
    m.discretizer.apply_to(m, nfe=3, wrt=m.fs.time, scheme="BACKWARD")

    # Initialisation - doen't really matter
    m.fs.tank.control_volume.material_holdup[:,:,:].fix(0.5)
    m.fs.tank.control_volume.material_holdup[:,:,:].unfix()

    # Setting the required properties

    #Inlet state - all timesteps
    m.fs.tank.inlet.flow_mol.fix(100)
    m.fs.tank.inlet.enth_mol.fix(m.fs.water.htpx(300 * pyo.units.K,101325* pyo.units.Pa)) # around 5 deg c
    m.fs.tank.inlet.pressure.fix(101325)

    #Tank properties
    m.fs.tank.tank_diameter.fix(0.5) # m
    m.fs.tank.heat_duty.fix(0)  # W
    m.fs.tank.outlet.flow_mol.fix(10)
    
    #Initial conditions
    m.fs.tank.initial_temperature.fix(300)
    m.fs.tank.initial_pressure.fix(101325)
    m.fs.tank.initial_holdup.fix(1000)

    assert degrees_of_freedom(m.fs) == 0

    m.fs.tank.initialize()


    opt = pyo.SolverFactory("ipopt")
    results = opt.solve(m, tee=False)
    #m.fs.tank.display()
    
    b = m.fs.tank

    assert results.solver.termination_condition == pyo.TerminationCondition.optimal
    
    # print("Enthalpy Liquid:   Enthalpy Vap:   Material Holdup Liquid: Material Holdup Vap")
    # for t in m.fs.time:
    #     print(m.fs.tank.enthalpy[t,"Liq"].value, m.fs.tank.enthalpy[t,"Vap"].value, pyo.value(m.fs.tank.material_phase_holdup[t,"Liq"]), pyo.value(m.fs.tank.material_phase_holdup[t,"Vap"]))

    # for t in m.fs.time:
    #     print(m.fs.tank.outlet.enth_mol[t].value, m.fs.tank.outlet.flow_mol[t].value, pyo.value(m.fs.tank.control_volume.properties_out[t].temperature), pyo.value(m.fs.tank.control_volume.properties_in[t].temperature))
        

    assert pyo.value(b.outlet.enth_mol[0]) == pytest.approx(pyo.value(b.inlet.enth_mol[0]), rel=1e-1)



if __name__ == "__main__":
    test_dynamic_tank()


from pydantic import BaseModel
from ahuora_builder_types.payloads.ml_request_schema import MLTrainRequestPayload, MLTrainingCompletionPayload
import pandas as pd
import json
import numpy as np
from io import StringIO
import contextlib
from sklearn.model_selection import train_test_split
from sklearn.metrics import mean_squared_error, r2_score
from idaes.core.surrogate.pysmo_surrogate import PysmoRBFTrainer, PysmoSurrogate


class MLResult(BaseModel):
    surrogate_model: dict
    charts: list[dict]
    metrics: list[dict]
    test_inputs: dict
    test_outputs: dict
    task_id: int


def ml_generate(schema: MLTrainRequestPayload) -> MLResult:
    df = pd.DataFrame(schema.datapoints, columns=schema.columns)
    train_df, test_df = train_test_split(df, test_size=0.2, random_state=42)

    input_labels = schema.input_labels
    output_labels = schema.output_labels
            
    trainer = PysmoRBFTrainer(
        input_labels=input_labels, output_labels=output_labels, training_dataframe=train_df)
    trainer.config.basis_function = 'gaussian'

    # Train surrogate (calls PySMO through IDAES Python wrapper)
    stream = StringIO()
    with contextlib.redirect_stdout(stream):
        rbf_train = trainer.train_surrogate()


    # create callable surrogate model
    rbf_surr = PysmoSurrogate(rbf_train, input_labels,
                              output_labels, input_bounds=None)
    f = StringIO()
    rbf_surr.save(f)
    content = f.getvalue()
    json_data = json.loads(content)

    df_evaluate = rbf_surr.evaluate_surrogate(test_df)

    metrics = []

    charts = []

    for output_label in output_labels:
        charts.append(compute_chart(
            test_df[output_label], df_evaluate[output_label], output_label))
        metrics.append({
            "mean_squared_error": round(mean_squared_error(df_evaluate[output_label], test_df[output_label]), 4),
            "r2_score": round(r2_score(df_evaluate[output_label], test_df[output_label]), 4),
        })

    return MLResult(
        surrogate_model=json_data,
        charts=charts,
        metrics=metrics,
        test_inputs=test_df.to_dict(orient='index'),
        test_outputs=df_evaluate.to_dict(orient='index'),
        task_id=schema.task_id
    )


def compute_chart(test_data_df, eval_data_df, output_label):
    minn = round(np.min([np.min(test_data_df), np.min(eval_data_df)]), 4)
    maxx = round(np.max([np.max(test_data_df), np.max(eval_data_df)]), 4)
    qq_plot_data = compute_qq_coordinates(test_data_df, eval_data_df)

    return {
        "min": minn,
        "max": maxx,
        "qq_plot_data": qq_plot_data,
        "output_label": output_label
    }


def compute_qq_coordinates(test_data, eval_data):
    """Compute QQ plot coordinates for two datasets."""
    test_values = test_data.to_numpy().flatten()
    eval_values = eval_data.to_numpy().flatten()

    # Sort values
    test_values.sort()
    eval_values.sort()

    # Generate QQ plot data points
    quantiles = np.linspace(0, 1, len(test_values))
    test_quantiles = np.quantile(test_values, quantiles)
    eval_quantiles = np.quantile(eval_values, quantiles)

    # Prepare JSON response for frontend
    qq_data = [{"x": round(float(t), 4), "y": round(float(e), 4)}
               for t, e in zip(test_quantiles, eval_quantiles)]
    return json.dumps(qq_data)



# Import Pyomo tools
from pyomo.environ import (
    Constraint,
    Var,
    Param,
    Expression,
    Reals,
    NonNegativeReals,
    Suffix,
)
from pyomo.environ import units as pyunits
from pyomo.common.config import ConfigBlock, ConfigValue, Bool

# Import IDAES cores
from idaes.core import (
    declare_process_block_class,
    MaterialFlowBasis,
    PhysicalParameterBlock,
    StateBlockData,
    StateBlock,
    MaterialBalanceType,
    EnergyBalanceType,
)
from idaes.core.base.components import Component
from idaes.core.base.phases import LiquidPhase
from idaes.core.util.initialization import (
    fix_state_vars,
    revert_state_vars,
    solve_indexed_blocks,
)
from idaes.core.base.process_base import ProcessBlockData
from idaes.core.base import property_meta
from idaes.core.util.model_statistics import (
    degrees_of_freedom,
    number_unfixed_variables,
)
from idaes.core.util.exceptions import PropertyPackageError
import idaes.core.util.scaling as iscale
import idaes.logger as idaeslog
from idaes.core.solvers import get_solver

# Set up logger
_log = idaeslog.getLogger(__name__)


# STEP 2

# When using this file the name "PowerParameterBlock" is what is imported
@declare_process_block_class("PowerParameterBlock")
class PowerParameterData(PhysicalParameterBlock):
    CONFIG = ProcessBlockData.CONFIG()
    CONFIG.declare(
            "default_arguments",
            ConfigBlock(
                implicit=True, description="Default arguments to use with Property Package"
            ),
        )

    def build(self):
        """
        Callable method for Block construction.
        """
        super(PowerParameterData, self).build()

        self._state_block_class = PowerStateBlock

        # Variables
        self.power = Var(initialize=0, domain=Reals, units=pyunits.W)

        # Default scaling values should be provided so that our tools can ensure the model is well-scaled
        # Generally scaling factors should be such that if it is multiplied by the variable it will range between 0.01 and 100
        self.set_default_scaling("power", 1e-3)

    @classmethod
    def define_metadata(cls, obj):
        # see https://github.com/watertap-org/watertap/blob/main/tutorials/creating_a_simple_property_model.ipynb
        obj.add_properties(
            {
                "power": {"method": None},
            }
        )
        obj.add_default_units(
            {
                "time": pyunits.s,
                "length": pyunits.m,
                "mass": pyunits.kg,
            }
        )
    
    def build_state_block(self, *args, **kwargs):
        """
        Methods to construct a StateBlock associated with this
        PhysicalParameterBlock. This will automatically set the parameters
        construction argument for the StateBlock.

        Returns:
            StateBlock

        """
        # default = kwargs.pop("default", {})
        initialize = kwargs.pop("initialize", {})

        if initialize == {}:
            kwargs["parameters"] = self
        else:
            for i in initialize.keys():
                initialize[i]["parameters"] = self

        return self.state_block_class(  # pylint: disable=not-callable
            *args, **kwargs, initialize=initialize
        )

    @property
    def state_block_class(self):
        if self._state_block_class is not None:
            return self._state_block_class
        else:
            raise AttributeError(
                "{} has not assigned a StateBlock class to be associated "
                "with this property package. Please contact the developer of "
                "the property package.".format(self.name)
            )

# STEP 3: State Block
class _PowerStateBlock(StateBlock):
    def initialize(
        self,
        state_args=None,
        state_vars_fixed=False,
        hold_state=False,
        outlvl=idaeslog.NOTSET,
        solver=None,
        optarg=None,
    ):
        """
        Initialization routine for property package.
        Keyword Arguments:
            state_args : Dictionary with initial guesses for the state vars
                         chosen. Note that if this method is triggered
                         through the control volume, and if initial guesses
                         were not provided at the unit model level, the
                         control volume passes the inlet values as initial
                         guess.The keys for the state_args dictionary are:

                         flow_mass_phase_comp : value at which to initialize
                                               phase component flows
                         pressure : value at which to initialize pressure
                         temperature : value at which to initialize temperature

            state_vars_fixed: Flag to denote if state vars have already been
                              fixed.
                              - True - states have already been fixed by the
                                       control volume 1D. Control volume 0D
                                       does not fix the state vars, so will
                                       be False if this state block is used
                                       with 0D blocks.
                             - False - states have not been fixed. The state
                                       block will deal with fixing/unfixing.
            hold_state : flag indicating whether the initialization routine
                         should unfix any state variables fixed during
                         initialization (default=False).
                         - True - states variables are not unfixed, and
                                 a dict of returned containing flags for
                                 which states were fixed during
                                 initialization.
                        - False - state variables are unfixed after
                                 initialization by calling the
                                 release_state method
            outlvl : sets output level of initialization routine (default=idaeslog.NOTSET)
            solver : Solver object to use during initialization if None is provided
                     it will use the default solver for IDAES (default = None)
            optarg : solver options dictionary object (default=None)
        Returns:
            If hold_states is True, returns a dict containing flags for
            which states were fixed during initialization.
        """

        # Fix state variables
        flags = fix_state_vars(self, state_args)
        # Check that dof = 0 when state variables are fixed
        for k in self.keys():
            dof = degrees_of_freedom(self[k])
            if dof != 0:
                raise PropertyPackageError(
                    "\nWhile initializing {sb_name}, the degrees of freedom "
                    "are {dof}, when zero is required. \nInitialization assumes "
                    "that the state variables should be fixed and that no other "
                    "variables are fixed. \nIf other properties have a "
                    "predetermined value, use the calculate_state method "
                    "before using initialize to determine the values for "
                    "the state variables and avoid fixing the property variables."
                    "".format(sb_name=self.name, dof=dof)
                )

        # If input block, return flags, else release state
        if state_vars_fixed is False:
            if hold_state is True:
                return flags
            else:
                self.release_state(flags)

    def release_state(self, flags, outlvl=idaeslog.NOTSET):
        """
        Method to release state variables fixed during initialisation.

        Keyword Arguments:
            flags : dict containing information of which state variables
                    were fixed during initialization, and should now be
                    unfixed. This dict is returned by initialize if
                    hold_state=True.
            outlvl : sets output level of of logging
        """
        if flags is None:
            return
        # Unfix state variables
        for attr in flags:
            if flags[attr] is True:
                getattr(self, attr).unfix()
        return
        
# STEP 4: 
@declare_process_block_class("PowerStateBlock", block_class=_PowerStateBlock)
class PowerStateBlockData(StateBlockData):
    def build(self):
        """Callable method for Block construction."""
        super(PowerStateBlockData, self).build()

        self.scaling_factor = Suffix(direction=Suffix.EXPORT)

        self.power = Var(
            initialize=0,
            domain=Reals,
            units=pyunits.W,
            doc="Power flow",
        )

    # -----------------------------------------------------------------------------

    def define_state_vars(self):
        """Define state vars."""
        return {
            "power": self.power,
        }

    # -----------------------------------------------------------------------------
    # Scaling methods
    def calculate_scaling_factors(self):
        super().calculate_scaling_factors()
        # This doesn't do anything, but it's a good example of how to get and set scaling factors in relation to each other.
        sf = iscale.get_scaling_factor(self.power)
        iscale.set_scaling_factor(self.power, sf)





#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Heat Exchanger Models.
"""

__author__ = "Team Ahuora"



# Import Pyomo libraries
from pyomo.environ import (
    Block,
    Var,
    Param,
    log,
    Reference,
    PositiveReals,
    ExternalFunction,
    units as pyunits,
    check_optimal_termination,
)
from pyomo.common.config import ConfigBlock, ConfigValue, In

# Import IDAES cores
from idaes.core import (
    declare_process_block_class,
    UnitModelBlockData,
)

import idaes.logger as idaeslog
from idaes.core.util.functions import functions_lib
from idaes.core.util.tables import create_stream_table_dataframe
from idaes.models.unit_models.heater import (
    _make_heater_config_block,
    _make_heater_control_volume,
)

from idaes.core.util.misc import add_object_reference
from idaes.core.util import scaling as iscale
from idaes.core.solvers import get_solver
from idaes.core.util.exceptions import ConfigurationError, InitializationError
from idaes.core.initialization import SingleControlVolumeUnitInitializer

_log = idaeslog.getLogger(__name__)


class SimpleHeatPumpInitializer(SingleControlVolumeUnitInitializer):
    """
    Initializer for 0D Heat Exchanger units.

    """

    def initialization_routine(
        self,
        model: Block,
        plugin_initializer_args: dict = None,
        copy_inlet_state: bool = False,
        duty=1000 * pyunits.W,
    ):
        """
        Common initialization routine for 0D Heat Exchangers.

        This routine starts by initializing the hot and cold side properties. Next, the heat
        transfer between the two sides is fixed to an initial guess for the heat duty (provided by the duty
        argument), the associated constraint is deactivated, and the model is then solved. Finally, the heat
        duty is unfixed and the heat transfer constraint reactivated followed by a final solve of the model.

        Args:
            model: Pyomo Block to be initialized
            plugin_initializer_args: dict-of-dicts containing arguments to be passed to plug-in Initializers.
                Keys should be submodel components.
            copy_inlet_state: bool (default=False). Whether to copy inlet state to other states or not
                (0-D control volumes only). Copying will generally be faster, but inlet states may not contain
                all properties required elsewhere.
            duty: initial guess for heat duty to assist with initialization. Can be a Pyomo expression with units.

        Returns:
            Pyomo solver results object
        """
        return super(SingleControlVolumeUnitInitializer, self).initialization_routine(
            model=model,
            plugin_initializer_args=plugin_initializer_args,
            copy_inlet_state=copy_inlet_state,
            duty=duty,
        )

    def initialize_main_model(
        self,
        model: Block,
        copy_inlet_state: bool = False,
        duty=1000 * pyunits.W,
    ):
        """
        Initialization routine for main 0D HX models.

        Args:
            model: Pyomo Block to be initialized.
            copy_inlet_state: bool (default=False). Whether to copy inlet state to other states or not
                (0-D control volumes only). Copying will generally be faster, but inlet states may not contain
                all properties required elsewhere.
            duty: initial guess for heat duty to assist with initialization, default = 1000 W. Can
                be a Pyomo expression with units.

        Returns:
            Pyomo solver results object.

        """
        # Get loggers
        init_log = idaeslog.getInitLogger(
            model.name, self.get_output_level(), tag="unit"
        )
        solve_log = idaeslog.getSolveLogger(
            model.name, self.get_output_level(), tag="unit"
        )

        # Create solver
        solver = self._get_solver()

        self.initialize_control_volume(model.source, copy_inlet_state)
        init_log.info_high("Initialization Step 1a (heat output (sink)) Complete.")

        self.initialize_control_volume(model.sink, copy_inlet_state)
        init_log.info_high("Initialization Step 1b (cold side) Complete.")
        # ---------------------------------------------------------------------
        # Solve unit without heat transfer equation
        model.heat_transfer_equation.deactivate()

        # Check to see if heat duty is fixed
        # We will assume that if the first point is fixed, it is fixed at all points
        if not model.sink.heat[model.flowsheet().time.first()].fixed:
            cs_fixed = False

            model.sink.heat.fix(duty)
            for i in model.source.heat:
                model.source.heat[i].set_value(-duty)
        else:
            cs_fixed = True
            for i in model.source.heat:
                model.source.heat[i].set_value(model.sink.heat[i])

        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = solver.solve(model, tee=slc.tee)
        init_log.info_high("Initialization Step 2 {}.".format(idaeslog.condition(res)))
        if not cs_fixed:
            model.sink.heat.unfix()
        model.heat_transfer_equation.activate()
        # ---------------------------------------------------------------------
        # Solve unit
        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = solver.solve(model, tee=slc.tee)
        init_log.info("Initialization Completed, {}".format(idaeslog.condition(res)))

        return res


def _make_heat_pump_config(config):
    """
    Declare configuration options for HeatExchangerData block.
    """
    config.declare(
        "source",
        ConfigBlock(
            description="Config block for heat output (sink)",
            doc="""A config block used to construct the heat output (sink) control volume.
This config can be given by the heat output (sink) name instead of source.""",
        ),
    )
    config.declare(
        "sink",
        ConfigBlock(
            description="Config block for cold side",
            doc="""A config block used to construct the cold side control volume.
This config can be given by the cold side name instead of sink.""",
        ),
    )
    _make_heater_config_block(config.source)
    _make_heater_config_block(config.sink)


@declare_process_block_class("SimpleHeatPump", doc="Simple 0D heat pump model.")
class SimpleHeatPumpData(UnitModelBlockData):
    """
    Simple 0D heat pump unit.
    Unit model to transfer heat from one material to another.
    """

    default_initializer = SimpleHeatPumpInitializer

    CONFIG = UnitModelBlockData.CONFIG(implicit=True)
    _make_heat_pump_config(CONFIG)

    def build(self):
        """
        Building model

        Args:
            None
        Returns:
            None
        """
        ########################################################################
        #  Call UnitModel.build to setup dynamics and configure                #
        ########################################################################
        super().build()
        config = self.config

        ########################################################################
        # Add control volumes                                                  #
        ########################################################################
        source = _make_heater_control_volume(
            self,
            "source",
            config.source,
            dynamic=config.dynamic,
            has_holdup=config.has_holdup,
        )
        sink = _make_heater_control_volume(
            self,
            "sink",
            config.sink,
            dynamic=config.dynamic,
            has_holdup=config.has_holdup,
        )

        ########################################################################
        # Add variables                                                        #
        ########################################################################
        # Use heat output (sink) units as basis
        s1_metadata = self.source.config.property_package.get_metadata()

        q_units = s1_metadata.get_derived_units("power")
        temp_units = s1_metadata.get_derived_units("temperature")

        self.work_mechanical = Var(
            self.flowsheet().time,
            domain=PositiveReals,
            initialize=100.0,
            doc="Mechanical work input",
            units=q_units,
        )
        self.coefficient_of_performance = Var(
            domain=PositiveReals,
            initialize=2.0,
            doc="Coefficient of performance",
            units=pyunits.dimensionless,
        )
        self.efficiency = Var(
            domain=PositiveReals,
            initialize=0.5,
            doc="Efficiency (Carnot = 1)",
            units=pyunits.dimensionless,
        )
        self.delta_temperature_lift = Var(
            self.flowsheet().time,
            domain=PositiveReals,
            initialize=10.0,
            doc="Temperature lift between source inlet and sink outlet",
            units=temp_units,
        )
        
        self.approach_temperature = Var(
            domain=PositiveReals,
            initialize=10.0,
            doc="Approach temperature of refrigerant",
            units=temp_units,
        )

        self.heat_duty = Reference(sink.heat)
        ########################################################################
        # Add ports                                                            #
        ########################################################################
        self.add_inlet_port(name="source_inlet", block=source, doc="heat output (sink) inlet")
        self.add_inlet_port(
            name="sink_inlet",
            block=sink,
            doc="Cold side inlet",
        )
        self.add_outlet_port(
            name="source_outlet", block=source, doc="heat output (sink) outlet"
        )
        self.add_outlet_port(
            name="sink_outlet",
            block=sink,
            doc="Cold side outlet",
        )

        ########################################################################
        # Add temperature lift constraints                                     #
        ########################################################################

        @self.Constraint(self.flowsheet().time)
        def delta_temperature_lift_equation(b, t):
            # Refrigerant saturation levels (K)
            T_cond = b.sink.properties_out[0].temperature + b.approach_temperature
            T_evap = b.source.properties_out[0].temperature - b.approach_temperature
            # Positive temperature lift
            return b.delta_temperature_lift[t] == T_cond - T_evap

        ########################################################################
        # Add a unit level energy balance                                      #
        ########################################################################
        @self.Constraint(self.flowsheet().time)
        def unit_heat_balance(b, t):
            # The duty of the sink = the source + work input
            return 0 == (
                source.heat[t] + pyunits.convert(sink.heat[t], to_units=q_units) - b.work_mechanical[t]
            )

        
        ########################################################################
        # Add Heat transfer equation                                           #
        ########################################################################

        @self.Constraint(self.flowsheet().time)
        def heat_transfer_equation(b, t):
            # This equation defines the heat duty using the cop of heating.
            return sink.heat[t] == b.coefficient_of_performance * b.work_mechanical[t]

        ########################################################################
        # Add COP Equation                                                     #
        ########################################################################
        @self.Constraint()
        def cop_equation(b):
            # Refrigerant saturation levels (K)
            T_cond = b.sink.properties_out[0].temperature + b.approach_temperature
            T_evap = b.source.properties_out[0].temperature - b.approach_temperature
            # Carnot CoP with efficiency
            return b.coefficient_of_performance == (T_cond / (T_cond - T_evap)) * b.efficiency
    
        
        ########################################################################
        # Add symbols for LaTeX equation rendering                             #
        ########################################################################
        self.work_mechanical.latex_symbol = "W_{mech}"
        self.coefficient_of_performance.latex_symbol = "COP"
        self.efficiency.latex_symbol = "\\eta"
        self.heat_duty.latex_symbol = "Q_{HP}"
        self.delta_temperature_lift.latex_symbol = "\\Delta T_{lift}"
     

    def initialize_build(
        self,
        state_args_1=None,
        state_args_2=None,
        outlvl=idaeslog.NOTSET,
        solver=None,
        optarg=None,
        duty=None,
    ):
        """
        Heat exchanger initialization method.

        Args:
            state_args_1 : a dict of arguments to be passed to the property
                initialization for the heat output (sink) (see documentation of the specific
                property package) (default = {}).
            state_args_2 : a dict of arguments to be passed to the property
                initialization for the cold side (see documentation of the specific
                property package) (default = {}).
            outlvl : sets output level of initialization routine
            optarg : solver options dictionary object (default=None, use
                     default solver options)
            solver : str indicating which solver to use during
                     initialization (default = None, use default solver)
            duty : an initial guess for the amount of heat transferred. This
                should be a tuple in the form (value, units), (default
                = (1000 J/s))

        Returns:
            None

        """
        # Set solver options
        init_log = idaeslog.getInitLogger(self.name, outlvl, tag="unit")
        solve_log = idaeslog.getSolveLogger(self.name, outlvl, tag="unit")

        # Create solver
        opt = get_solver(solver, optarg)

        flags1 = self.source.initialize(
            outlvl=outlvl, optarg=optarg, solver=solver, state_args=state_args_1
        )

        init_log.info_high("Initialization Step 1a (heat output (sink)) Complete.")

        flags2 = self.sink.initialize(
            outlvl=outlvl, optarg=optarg, solver=solver, state_args=state_args_2
        )

        init_log.info_high("Initialization Step 1b (cold side) Complete.")
        # ---------------------------------------------------------------------
        # Solve unit without heat transfer equation
        self.heat_transfer_equation.deactivate()

        # Get side 1 and side 2 heat units, and convert duty as needed
        s1_units = self.source.heat.get_units()
        s2_units = self.sink.heat.get_units()

        # Check to see if heat duty is fixed
        # WE will assume that if the first point is fixed, it is fixed at all points
        if not self.sink.heat[self.flowsheet().time.first()].fixed:
            cs_fixed = False
            if duty is None:
                # Assume 1000 J/s and check for unitless properties
                if s1_units is None and s2_units is None:
                    # Backwards compatibility for unitless properties
                    s1_duty = -1000
                    s2_duty = 1000
                else:
                    s1_duty = pyunits.convert_value(
                        -1000, from_units=pyunits.W, to_units=s1_units
                    )
                    s2_duty = pyunits.convert_value(
                        1000, from_units=pyunits.W, to_units=s2_units
                    )
            else:
                # Duty provided with explicit units
                s1_duty = -pyunits.convert_value(
                    duty[0], from_units=duty[1], to_units=s1_units
                )
                s2_duty = pyunits.convert_value(
                    duty[0], from_units=duty[1], to_units=s2_units
                )

            self.sink.heat.fix(s2_duty)
            for i in self.source.heat:
                self.source.heat[i].value = s1_duty
        else:
            cs_fixed = True
            for i in self.source.heat:
                self.source.heat[i].set_value(self.sink.heat[i])

        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = opt.solve(self, tee=slc.tee)
        init_log.info_high("Initialization Step 2 {}.".format(idaeslog.condition(res)))
        if not cs_fixed:
            self.sink.heat.unfix()
        self.heat_transfer_equation.activate()
        # ---------------------------------------------------------------------
        # Solve unit
        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = opt.solve(self, tee=slc.tee)
        init_log.info_high("Initialization Step 3 {}.".format(idaeslog.condition(res)))
        # ---------------------------------------------------------------------
        # Release Inlet state
        self.source.release_state(flags1, outlvl=outlvl)
        self.sink.release_state(flags2, outlvl=outlvl)

        init_log.info("Initialization Completed, {}".format(idaeslog.condition(res)))

        if not check_optimal_termination(res):
            raise InitializationError(
                f"{self.name} failed to initialize successfully. Please check "
                f"the output logs for more information."
            )

    def _get_performance_contents(self, time_point=0):
        # this shows as the performance table when calling the report method
        var_dict = {
            
            "Coefficient of Performance": self.coefficient_of_performance,
            "Efficiency": self.efficiency,
            "Source Heat Duty": self.source.heat[time_point],
            "Mechanical Work Input": self.work_mechanical[time_point],
            "Sink Heat Duty": self.sink.heat[time_point],
            "Temperature Lift": self.delta_temperature_lift[time_point],
        }

        expr_dict = {}

        return {"vars": var_dict, "exprs": expr_dict}

    def _get_stream_table_contents(self, time_point=0):
        # this shows as t he stream table when calling the report method
        # Get names for hot and cold sides
        return create_stream_table_dataframe(
            {
                f"Source Inlet": self.source_inlet,
                f"Source Outlet": self.source_outlet,
                f"Sink Inlet": self.sink_inlet,
                f"Sink Outlet": self.sink_outlet,
            },
            time_point=time_point,
        )

    def calculate_scaling_factors(self):
        super().calculate_scaling_factors()
        # TODO: Review this code to check it makes sense.

        # Scaling for heat pump variables
        # Mechanical work input: typical values vary, set default scaling to 0.01
        sf_work = dict(
            zip(
            self.work_mechanical.keys(),
            [
                iscale.get_scaling_factor(v, default=0.01)
                for v in self.work_mechanical.values()
            ],
            )
        )
        # COP and efficiency are dimensionless, usually between 1 and 10, default scaling 0.1
        sf_cop = iscale.get_scaling_factor(self.coefficient_of_performance, default=0.1)
        sf_eff = iscale.get_scaling_factor(self.efficiency, default=0.1)

        # Delta Ts: typical range 1-100, default scaling 0.1
        sf_dT_in = dict(
            zip(
            self.delta_temperature_in.keys(),
            [
                iscale.get_scaling_factor(v, default=0.1)
                for v in self.delta_temperature_in.values()
            ],
            )
        )
        sf_dT_out = dict(
            zip(
            self.delta_temperature_out.keys(),
            [
                iscale.get_scaling_factor(v, default=0.1)
                for v in self.delta_temperature_out.values()
            ],
            )
        )

        # Heat duty: depends on process, default scaling 0.01
        sf_q = dict(
            zip(
            self.heat_duty.keys(),
            [
                iscale.get_scaling_factor(v, default=0.01)
                for v in self.heat_duty.values()
            ],
            )
        )

        # Apply scaling to constraints
        for t, c in self.heat_transfer_equation.items():
            iscale.constraint_scaling_transform(
            c, sf_cop * sf_work[t], overwrite=False
            )

        for t, c in self.unit_heat_balance.items():
            iscale.constraint_scaling_transform(
            c, sf_q[t], overwrite=False
            )

        for t, c in self.delta_temperature_in_equation.items():
            iscale.constraint_scaling_transform(c, sf_dT_in[t], overwrite=False)

        for t, c in self.delta_temperature_out_equation.items():
            iscale.constraint_scaling_transform(c, sf_dT_out[t], overwrite=False)

        # COP equation scaling
        iscale.constraint_scaling_transform(
            self.cop_equation, sf_cop, overwrite=False
        )


from ahuora_builder_types.id_types import PropertyValueId
from pyomo.core.base.indexed_component import IndexedComponent
class PropertyComponent:
    def __init__(self, name: str, component : IndexedComponent, unknown_units=False):
        """
        Args:
        - name (str): The name of the property component (for debugging purposes)
        - component (Component): The Pyomo component; i.e., Expression, Var
        - unknown_units (bool): If units are unknown, idaes_factory will do some
            additional processing to determine the unit category.
        - corresponding_constraint (Constraint | Var | None): The component
            (Var or Constraint) that was fixed or activated to set the value of
            this property. Given an id in the properties dictionary, this field
            can be used to unfix or deactivate the corresponding constraint.
            Defaults to None (allowed for properties that are not fixed).
        """
        self.name: str = name
        self.component: IndexedComponent = component
        self.unknown_units = unknown_units
        self.corresponding_constraint = None # TODO: Type


class PropertiesManager:
    def __init__(self):
        self.properties : dict[PropertyValueId,PropertyComponent] = {}

    def add(self, id: PropertyValueId, indexed_component: IndexedComponent, name : str, unknown_units=False)-> IndexedComponent:
        self.properties[id] = PropertyComponent(name, indexed_component, unknown_units)
    
    def get(self, id: PropertyValueId):
        return self.properties[id]
    
    def get_component(self, id: PropertyValueId):
        return self.get(id).component
    
    def get_constraint(self, id: PropertyValueId):
        return self.get(id).corresponding_constraint
    
    def add_constraint(self, id: PropertyValueId, constraint):
        # assumes the property has already been added
        self.get(id).corresponding_constraint = constraint
    
    def items(self):
        return self.properties.items()


from idaes.models.control.controller import PIDControllerData, ControllerType, ControllerMVBoundType, ControllerAntiwindupType, smooth_bound, smooth_heaviside
from pyomo.environ import Var
from idaes.core import UnitModelBlockData, declare_process_block_class
from pyomo.common.config import ConfigValue, In, Bool
import pyomo.environ as pyo
from idaes.core.util.exceptions import ConfigurationError
import pyomo.dae as pyodae
from idaes.core.util import scaling as iscale
import functools


@declare_process_block_class(
    "PIDController2",
    doc="PID controller model block.  To use this the model must be dynamic.",
)
class PIDController2Data(UnitModelBlockData):
    """
    PID controller class.
    """

    CONFIG = UnitModelBlockData.CONFIG()
    CONFIG.declare(
        "mv_bound_type",
        ConfigValue(
            default=ControllerMVBoundType.NONE,
            domain=In(
                [
                    ControllerMVBoundType.NONE,
                    ControllerMVBoundType.SMOOTH_BOUND,
                    ControllerMVBoundType.LOGISTIC,
                ]
            ),
            description="Type of bounds to apply to the manipulated variable (mv)).",
            doc=(
                """Type of bounds to apply to the manipulated variable output. If,
                bounds are applied, the model parameters **mv_lb** and **mv_ub** set the bounds.
                The **default** is ControllerMVBoundType.NONE. See the controller documentation
                for details on the mathematical formulation. The options are:
                **ControllerMVBoundType.NONE** no bounds, **ControllerMVBoundType.SMOOTH_BOUND**
                smoothed mv = min(max(mv_unbound, ub), lb), and **ControllerMVBoundType.LOGISTIC**
                logistic function to enforce bounds.
                """
            ),
        ),
    )
    CONFIG.declare(
        "calculate_initial_integral",
        ConfigValue(
            default=True,
            domain=Bool,
            description="Calculate the initial integral term value if True",
            doc="Calculate the initial integral term value if True",
        ),
    )
    CONFIG.declare(
        "controller_type",
        ConfigValue(
            default=ControllerType.PI,
            domain=In(
                [
                    ControllerType.P,
                    ControllerType.PI,
                    ControllerType.PD,
                    ControllerType.PID,
                ]
            ),
            description="Control type",
            doc="""Controller type. The **default** = ControllerType.PI and the
                options are: **ControllerType.P** Proportional, **ControllerType.PI**
                proportional and integral, **ControllerType.PD** proportional and derivative, and
                **ControllerType.PID** proportional, integral, and derivative
                """,
        ),
    )
    CONFIG.declare(
        "antiwindup_type",
        ConfigValue(
            default=ControllerAntiwindupType.NONE,
            domain=In(
                [
                    ControllerAntiwindupType.NONE,
                    ControllerAntiwindupType.CONDITIONAL_INTEGRATION,
                    ControllerAntiwindupType.BACK_CALCULATION,
                ]
            ),
            description="Type of antiwindup technique to use.",
            doc=(
                """Type of antiwindup technique to use. Options are **ControllerAntiwindupType.NONE**,
                **ControllerAntiwindupType.CONDITIONAL_INTEGRATION**, and **ControllerAntiwindupType.BACK_CALCULATION**.
                See the controller documentation for details on the mathematical formulation.
                """
            ),
        ),
    )
    CONFIG.declare(
        "derivative_on_error",
        ConfigValue(
            default=False,
            domain=Bool,
            description="Whether basing derivative action on process var or error",
            doc="""Naive implementations of derivative action can cause large spikes in
                control when the setpoint is changed. One solution is to use the (negative)
                derivative of the process variable to calculate derivative action instead
                of using the derivative of setpoint error. If **True**, use the derivative of
                setpoint error to calculate derivative action. If **False** (default), use the
                (negative) derivative of the process variable instead.
                """,
        ),
    )

    def build(self):
        """
        Build the PID block
        """
        super().build()


        if self.config.dynamic is False:
            raise ConfigurationError(
                "PIDControllers work only with dynamic flowsheets."
            )

        """
        We want to create the controlled and manipulated variables internally, and then the user can use constraints to set them later on.
        """

        self.process_var = Var(
            self.flowsheet().time,
            initialize=0.0,
            doc="Setpoint for the PID controller",
        )
        self.manipulated_var = Var(
            self.flowsheet().time,
            initialize=0.0,
            doc="Manipulated variable for the PID controller",
        )

        # Shorter pointers to time set information
        time_set = self.flowsheet().time
        time_units = self.flowsheet().time_units
        if time_units is None:
            time_units = pyo.units.dimensionless
        t0 = time_set.first()

        # Type Check
        if not issubclass(self.process_var[t0].ctype, (pyo.Var, pyo.Expression)):
            raise TypeError(
                f"process_var must reference a Var or Expression not {self.process_var[t0].ctype}"
            )
        if not issubclass(self.manipulated_var[t0].ctype, pyo.Var):
            raise TypeError(
                f"manipulated_var must reference a Var not {self.manipulated_var[t0].ctype}"
            )

        if not self.config.antiwindup_type == ControllerAntiwindupType.NONE:
            if not self.config.controller_type in [
                ControllerType.PI,
                ControllerType.PID,
            ]:
                raise ConfigurationError(
                    "User specified antiwindup method for controller without integral action."
                )
            if self.config.mv_bound_type == ControllerMVBoundType.NONE:
                raise ConfigurationError(
                    "User specified antiwindup method for unbounded MV."
                )

        # Get the appropriate units for various controller variables
        mv_units = pyo.units.get_units(self.manipulated_var[t0])
        pv_units = pyo.units.get_units(self.process_var[t0])
        if mv_units is None:
            mv_units = pyo.units.dimensionless
        if pv_units is None:
            pv_units = pyo.units.dimensionless
        gain_p_units = mv_units / pv_units

        if not self.config.mv_bound_type == ControllerMVBoundType.NONE:
            # Parameters
            self.mv_lb = pyo.Param(
                mutable=True,
                initialize=0.05,
                doc="Controller output lower bound",
                units=mv_units,
            )
            self.mv_ub = pyo.Param(
                mutable=True,
                initialize=1,
                doc="Controller output upper bound",
                units=mv_units,
            )
        if self.config.mv_bound_type == ControllerMVBoundType.SMOOTH_BOUND:
            self.smooth_eps = pyo.Param(
                mutable=True,
                initialize=1e-4,
                doc="Smoothing parameter for controller output limits when the bound"
                " type is SMOOTH_BOUND",
                units=mv_units,
            )
        elif self.config.mv_bound_type == ControllerMVBoundType.LOGISTIC:
            self.logistic_bound_k = pyo.Param(
                mutable=True,
                initialize=4,
                doc="Smoothing parameter for controller output limits when the bound"
                " type is LOGISTIC",
                units=pyo.units.dimensionless,
            )
        if (
            self.config.antiwindup_type
            == ControllerAntiwindupType.CONDITIONAL_INTEGRATION
        ):
            self.conditional_integration_k = pyo.Param(
                mutable=True,
                initialize=200,
                doc="Parameter governing steepness of transition between integrating and not integrating."
                "A larger value means a steeper transition.",
                units=pyo.units.dimensionless,
            )

        # Variable for basic controller settings may change with time.
        self.setpoint = pyo.Var(
            time_set, initialize=0.5, doc="Setpoint", units=pv_units
        )
        self.gain_p = pyo.Var(
            time_set,
            initialize=0.1,
            doc="Gain for proportional part",
            units=gain_p_units,
        )
        if self.config.controller_type in [ControllerType.PI, ControllerType.PID]:
            self.gain_i = pyo.Var(
                time_set,
                initialize=0.1,
                doc="Gain for integral part",
                units=gain_p_units / time_units,
            )
        if self.config.controller_type in [ControllerType.PD, ControllerType.PID]:
            self.gain_d = pyo.Var(
                time_set,
                initialize=0.01,
                doc="Gain for derivative part",
                units=gain_p_units * time_units,
            )

        if self.config.antiwindup_type == ControllerAntiwindupType.BACK_CALCULATION:
            self.gain_b = pyo.Var(
                time_set,
                initialize=0.1,
                doc="Gain for back calculation antiwindup",
                units=1 / time_units,
            )

        self.mv_ref = pyo.Var(
            time_set,
            initialize=0.5,
            doc="Controller bias",
            units=mv_units,
        )

        # Error expression or variable (variable required for derivative term)
        if (
            self.config.controller_type in [ControllerType.PD, ControllerType.PID]
            and self.config.derivative_on_error
        ):
            self.error = pyo.Var(
                time_set, initialize=0, doc="Error variable", units=pv_units
            )

            @self.Constraint(time_set, doc="Error constraint")
            def error_eqn(b, t):
                return b.error[t] == b.setpoint[t] - b.process_var[t]

            self.derivative_term = pyodae.DerivativeVar(
                self.error,
                wrt=self.flowsheet().time,
                initialize=0,
                units=pv_units / time_units,
            )

        else:

            @self.Expression(time_set, doc="Error expression")
            def error(b, t):
                return b.setpoint[t] - b.process_var[t]

            if (
                self.config.controller_type in [ControllerType.PD, ControllerType.PID]
                and not self.config.derivative_on_error
            ):
                # Need to create a Var because process_var might be an Expression
                self.negative_pv = pyo.Var(
                    time_set,
                    initialize=0,
                    doc="Negative of process variable",
                    units=pv_units,
                )

                @self.Constraint(time_set, doc="Negative process variable equation")
                def negative_pv_eqn(b, t):
                    return b.negative_pv[t] == -b.process_var[t]

                self.derivative_term = pyodae.DerivativeVar(
                    self.negative_pv,
                    wrt=self.flowsheet().time,
                    initialize=0,
                    units=pv_units / time_units,
                )

        # integral term written de_i(t)/dt = e(t)
        if self.config.controller_type in [ControllerType.PI, ControllerType.PID]:
            self.mv_integral_component = pyo.Var(
                time_set,
                initialize=0,
                doc="Integral contribution to control action",
                units=mv_units,
            )
            self.mv_integral_component_dot = pyodae.DerivativeVar(
                self.mv_integral_component,
                wrt=time_set,
                initialize=0,
                units=mv_units / time_units,
                doc="Rate of change of integral contribution to control action",
            )

            if self.config.calculate_initial_integral:

                @self.Constraint(doc="Calculate initial e_i based on output")
                def initial_integral_error_eqn(b):
                    if self.config.controller_type == ControllerType.PI:
                        return b.mv_integral_component[t0] == (
                            b.manipulated_var[t0]
                            - b.mv_ref[t0]
                            - b.gain_p[t0] * b.error[t0]
                        )
                    return b.mv_integral_component[t0] == (
                        b.manipulated_var[t0]
                        - b.mv_ref[t0]
                        - b.gain_p[t0] * b.error[t0]
                        - b.gain_d[t0] * b.derivative_term[t0]
                    )

        @self.Expression(time_set, doc="Unbounded output for manipulated variable")
        def mv_unbounded(b, t):
            if self.config.controller_type == ControllerType.PID:
                return (
                    b.mv_ref[t]
                    + b.gain_p[t] * b.error[t]
                    + b.mv_integral_component[t]
                    + b.gain_d[t] * b.derivative_term[t]
                )
            elif self.config.controller_type == ControllerType.PI:
                return (
                    b.mv_ref[t] + b.gain_p[t] * b.error[t] + b.mv_integral_component[t]
                )
            elif self.config.controller_type == ControllerType.PD:
                return (
                    b.mv_ref[t]
                    + b.gain_p[t] * b.error[t]
                    + b.gain_d[t] * b.derivative_term[t]
                )
            elif self.config.controller_type == ControllerType.P:
                return b.mv_ref[t] + b.gain_p[t] * b.error[t]
            else:
                raise ConfigurationError(
                    f"{self.config.controller_type} is not a valid PID controller type"
                )

        @self.Constraint(time_set, doc="Bounded output of manipulated variable")
        def mv_eqn(b, t):
            if self.config.mv_bound_type == ControllerMVBoundType.SMOOTH_BOUND:
                return b.manipulated_var[t] == smooth_bound(
                    b.mv_unbounded[t], lb=b.mv_lb, ub=b.mv_ub, eps=b.smooth_eps
                )
            elif self.config.mv_bound_type == ControllerMVBoundType.LOGISTIC:
                return (
                    (b.manipulated_var[t] - b.mv_lb)
                    * (
                        1
                        + pyo.exp(
                            -b.logistic_bound_k
                            * (b.mv_unbounded[t] - (b.mv_lb + b.mv_ub) / 2)
                            / (b.mv_ub - b.mv_lb)
                        )
                    )
                ) == b.mv_ub - b.mv_lb
            return b.manipulated_var[t] == b.mv_unbounded[t]

        # deactivate the time 0 mv_eqn instead of skip, should be fine since
        # first time step always exists.
        if self.config.calculate_initial_integral:
            self.mv_eqn[t0].deactivate()

        if self.config.controller_type in [ControllerType.PI, ControllerType.PID]:

            @self.Constraint(time_set, doc="de_i(t)/dt = e(t)")
            def mv_integration_eqn(b, t):
                if (
                    self.config.antiwindup_type
                    == ControllerAntiwindupType.CONDITIONAL_INTEGRATION
                ):
                    # This expression is not sensitive to whether the "right" or "wrong" bound is active for a given
                    # expression of error.
                    return b.mv_integral_component_dot[t] == b.gain_i[t] * b.error[
                        t
                    ] * (
                        smooth_heaviside(
                            (b.mv_unbounded[t] - b.mv_lb) / (b.mv_ub - b.mv_lb),
                            b.conditional_integration_k,
                        )
                        # 1
                        - smooth_heaviside(
                            (b.mv_unbounded[t] - b.mv_ub) / (b.mv_ub - b.mv_lb),
                            b.conditional_integration_k,
                        )
                    )
                elif (
                    self.config.antiwindup_type
                    == ControllerAntiwindupType.BACK_CALCULATION
                ):
                    return b.mv_integral_component_dot[t] == b.gain_i[t] * b.error[
                        t
                    ] + b.gain_b[t] * (b.manipulated_var[t] - b.mv_unbounded[t])
                else:
                    return b.mv_integral_component_dot[t] == b.gain_i[t] * b.error[t]

            self.mv_integration_eqn[t0].deactivate()


            # TODO: Initial values.
            # For now, mv_integral_component is 0
            self.mv_integral_component[t0].fix(0)

    def calculate_scaling_factors(self):
        super().calculate_scaling_factors()
        gsf = iscale.get_scaling_factor
        ssf = functools.partial(iscale.set_scaling_factor, overwrite=False)
        cst = functools.partial(iscale.constraint_scaling_transform, overwrite=False)

        # orig_pv = self.config.process_var
        # orig_mv = self.config.manipulated_var
        time_set = self.flowsheet().time
        t0 = time_set.first()

        sf_pv = iscale.get_scaling_factor(self.config.process_var[t0])
        if sf_pv is None:
            sf_pv = iscale.get_scaling_factor(self.process_var[t0], default=1)
        sf_mv = iscale.get_scaling_factor(self.config.manipulated_var[t0])
        if sf_mv is None:
            sf_mv = iscale.get_scaling_factor(self.manipulated_var[t0], default=1)

        # Don't like calling scaling laterally like this, but we need scaling factors for the pv and mv
        # Except this causes a StackOverflow with flowsheet-level PVs or MVs---put this on ice for now
        # if sf_pv is None:
        #     try:
        #         iscale.calculate_scaling_factors(self.config.process_var.parent_block())
        #     except RecursionError:
        #         raise ConfigurationError(
        #             f"Circular scaling dependency detected in Controller {self.name}. The only way this should be "
        #             "able to happen is if a loop of controllers exists manipulating each others setpoints without "
        #             "terminating in an actual process variable."
        #         )
        # if sf_mv is None:
        #     try:
        #         iscale.calculate_scaling_factors(self.config.manipulated_var.parent_block())
        #     except RecursionError:
        #         raise ConfigurationError(
        #             f"Circular scaling dependency detected in Controller {self.name}. The only way this should be "
        #             "able to happen is if a loop of controllers exists manipulating each others setpoints without "
        #             "terminating in an actual process variable."
        #         )

        if self.config.calculate_initial_integral:
            sf_mv = gsf(self.manipulated_var[t0], default=1, warning=True)
            cst(self.initial_integral_error_eqn, sf_mv)

        for t in time_set:
            sf_pv = gsf(self.process_var[t], default=1, warning=True)
            sf_mv = gsf(self.manipulated_var[t], default=1, warning=True)

            ssf(self.setpoint[t], sf_pv)
            ssf(self.mv_ref[t], sf_mv)
            cst(self.mv_eqn[t], sf_mv)

            if self.config.controller_type in [ControllerType.PD, ControllerType.PID]:
                if self.config.derivative_on_error:
                    ssf(self.error[t], sf_pv)
                    cst(self.error_eqn[t], sf_pv)
                else:
                    ssf(self.negative_pv[t], sf_pv)
                    cst(self.negative_pv_eqn[t], sf_pv)

            if self.config.controller_type in [ControllerType.PI, ControllerType.PID]:
                ssf(self.mv_integral_component[t], sf_mv)

                cst(self.mv_integration_eqn[t], sf_pv)



from pyomo.environ import Reference, Var, units
from pyomo.dae import DerivativeVar


def add_initial_dynamics(unit_model):
    """
    Adds the reference variables for initial holdup and initial accumulation. 
    Makes it easier for us to set initial conditions in the frontend, as we can reference them directly.
    """
    if unit_model.config.dynamic:
            # add initial holdup reference
            unit_model.initial_material_holdup = Reference(unit_model.control_volume.material_holdup[0,:,:])
            unit_model.initial_energy_holdup = Reference(unit_model.control_volume.energy_holdup[0,:])

            # For some reason we can't do references to the initialaccumulation variables,
            # Error ( Can only take the derivative of a Varcomponent.)
            #  so we create them as vars
            unit_model.initial_material_accumulation = Var(unit_model.config.property_package.phase_list, unit_model.config.property_package.component_list, initialize=0,units=units.mol/units.s)
            unit_model.initial_energy_accumulation = Var(unit_model.config.property_package.phase_list, initialize=0,units=units.kW)


            @unit_model.Constraint(
                unit_model.config.property_package.phase_list,
                unit_model.config.property_package.component_list,
                doc="Initial material accumulation constraint"
            )
            def initial_material_accumulation_constraint(b, p, j):
                return b.initial_material_accumulation[p, j] == b.control_volume.material_accumulation[0, p, j]
            
            @unit_model.Constraint(
                unit_model.config.property_package.phase_list,
                doc="Initial energy accumulation constraint"
            )
            def initial_energy_accumulation_constraint(b, p):
                return b.initial_energy_accumulation[p] == b.control_volume.energy_accumulation[0, p]


# Import Pyomo libraries
from pyomo.environ import (
    Var,
    Suffix,
    units as pyunits,
)
from pyomo.common.config import ConfigBlock, ConfigValue, In
from idaes.core.util.tables import create_stream_table_dataframe
from idaes.core.util.exceptions import ConfigurationError
# Import IDAES cores
from idaes.core import (
    declare_process_block_class,
    UnitModelBlockData,
    useDefault,
)
from idaes.core.util.config import is_physical_parameter_block
import idaes.core.util.scaling as iscale
import idaes.logger as idaeslog

from ..custom.updated_pressure_changer import (

    PumpData,
)




# When using this file the name "CustomCompressor" is what is imported
@declare_process_block_class("CustomPump")
class CustomPumpData(PumpData):
    """
    Zero order Load model
    """

    # CONFIG are options for the unit model, this simple model only has the mandatory config options
    CONFIG = PumpData.CONFIG()

    CONFIG.declare(
        "power_property_package",
        ConfigValue(
            default=useDefault,
            domain=is_physical_parameter_block,
            description="Property package to use for power",
            doc="""Power Property parameter object used to define power calculations,
    **default** - useDefault.
    **Valid values:** {
    **useDefault** - use default package from parent model or flowsheet,
    **PhysicalParameterObject** - a PhysicalParameterBlock object.}""",
        ),
    )
    CONFIG.declare(
        "power_property_package_args",
        ConfigBlock(
            implicit=True,
            description="Arguments to use for constructing power property packages",
            doc="""A ConfigBlock with arguments to be passed to a property block(s)
    and used when constructing these,
    **default** - None. 
    **Valid values:** {
    see property package for documentation.}""",
        ),
    )

    def build(self):
        # build always starts by calling super().build()
        # This triggers a lot of boilerplate in the background for you
        super().build()

        # This creates blank scaling factors, which are populated later
        self.scaling_factor = Suffix(direction=Suffix.EXPORT)


        # Add state blocks for inlet, outlet, and waste
        # These include the state variables and any other properties on demand
 
        tmp_dict = dict(**self.config.property_package_args)
        tmp_dict["parameters"] = self.config.property_package
        tmp_dict["defined_state"] = True  # inlet block is an inlet
       # Add inlet block
        # self.properties_in = self.config.property_package.state_block_class(
        #     self.flowsheet().config.time, doc="Material properties of inlet", **tmp_dict
        # )


        # Add outlet and waste block
        tmp_dict["defined_state"] = False  # outlet and waste block is not an inlet
        self.power_properties_out = self.config.power_property_package.state_block_class(
            self.flowsheet().config.time,
            doc="Material properties of outlet",
            **tmp_dict
        )

        # Add ports - oftentimes users interact with these rather than the state blocks
        self.add_port(name="power_outlet", block=self.power_properties_out)

        # Add constraints
        # Usually unit models use a control volume to do the mass, energy, and momentum
        # balances, however, they will be explicitly written out in this example
        @self.Constraint(
            self.flowsheet().time,
            doc="Power out",
        )
        def eq_power_out(b, t):
            return (
                self.power_properties_out[t].power == self.work_mechanical[t] * -1
            )
    


### Imports
from pyomo.environ import ConcreteModel, SolverFactory, SolverStatus, TerminationCondition, Block, TransformationFactory, assert_optimal_termination
from pyomo.network import SequentialDecomposition, Port, Arc
from pyomo.core.base.units_container import _PyomoUnit, units 
from idaes.core import FlowsheetBlock
from idaes.core.util.model_statistics import report_statistics, degrees_of_freedom
from idaes.core.util.tables import _get_state_from_port
import idaes.logger as idaeslog
from property_packages.build_package import build_package
from idaes.models.unit_models.pressure_changer import Pump
from idaes.models.unit_models.heater import Heater
from ..energy_mixer import EnergyMixer
from ..power_property_package import PowerParameterBlock
from pyomo.environ import value


def test_energy_mixer():
    m = ConcreteModel()
    m.fs = FlowsheetBlock()
    m.fs.power_property_package = PowerParameterBlock()

    m.fs.mixer = EnergyMixer(
        property_package=m.fs.power_property_package,
        num_inlets=3,
        )
        
    
    

    m.fs.mixer.inlet_1.power.fix(100 * units.kW) 
    m.fs.mixer.inlet_2.power.fix(1000 * units.W)
    m.fs.mixer.inlet_3.power.fix(10 * units.kW)
    m.fs.mixer.efficiency.fix(0.90)

    # Check that the model is fully specified
    assert degrees_of_freedom(m) == 0 

    # check that it solves correctly

    solver = SolverFactory("ipopt")
    solver.solve(m)
    #assert m.fs.mixer.outlet.power[0].value == 111 * units.kW
    assert abs(m.fs.mixer.outlet.power[0].value - 999e2) < 1e-3
    





from pyomo.environ import ConcreteModel, Var, value, Reference
from ..expression_parsing import parse_expression
from ...properties_manager import PropertiesManager


def test_parse_expression():
    m = ConcreteModel()
    m.x = Var(initialize=2)
    m.properties_map = PropertiesManager()
    m.properties_map.add(
        123, Reference(m.x), "m.x"
    )

    expr = parse_expression("id_123 + 2", m,1)
    assert value(expr) == 4



import pytest

# Import Pyomo libraries
import pyomo.environ as pyo
from pyomo.environ import ConcreteModel, SolverFactory, SolverStatus, TerminationCondition, Block, TransformationFactory, units, Objective, value, Constraint, Var
from pyomo.network import SequentialDecomposition, Port, Arc


# Import IDAES libraries
from idaes.core import FlowsheetBlock
from idaes.core.util.model_statistics import report_statistics, degrees_of_freedom
import idaes.logger as idaeslog
from idaes.core.util.tables import _get_state_from_port  

# Import required models
from idaes.models.unit_models import (
    Feed,
    Mixer,
    Heater,
    Compressor,
    Product,
    MomentumMixingType,
)
from idaes.models.unit_models import Separator as Splitter
from idaes.models.unit_models import Compressor, PressureChanger
from idaes.models.properties.general_helmholtz import (
    HelmholtzParameterBlock,
    PhaseType,
    StateVars,
    AmountBasis,
    )
from idaes.models.unit_models.pressure_changer import ThermodynamicAssumption, Turbine
from ahuora_builder.custom.thermal_utility_systems.willans_turbine import TurbineBase
from idaes.core.util import DiagnosticsToolbox

from pyomo.util.check_units import assert_units_consistent

@pytest.fixture
def build_model_with_inputs():
    def _make_case(calculation_method):
        # Define model components and blocks
        m = ConcreteModel()
        m.fs1 = FlowsheetBlock(dynamic=False)
        m.fs1.water = HelmholtzParameterBlock(
                        pure_component="h2o",
                        phase_presentation=PhaseType.LG,
                        state_vars=StateVars.PH,
                        amount_basis=AmountBasis.MOLE,
                        )
        
        m.fs1.turbine = TurbineBase(
            property_package=m.fs1.water, 
            calculation_method=calculation_method,
            )

        # Inputs
        n_in = 187.0 * 3.6 * 15.419020010014712 # t/h to mol/s
        P_in = 41.3 # bar g
        if m.fs1.turbine.config.calculation_method == "CT_willans": # CT model needs lower pressure
            P_out = -0.4 # bar g
        else:
            P_out = 10.4 # bar g
        T_in = 381 # C

        m.fs1.turbine.inlet.flow_mol.fix(n_in/3.6)
        m.fs1.turbine.inlet.enth_mol[0].fix(
            value(
                m.fs1.water.htpx(
                    T=(T_in+273)*units.K, 
                    p=(P_in+1)*units.bar,
                )
            )
        )
        m.fs1.turbine.inlet.pressure[0].fix((P_in+1)*units.bar) # why the [0]
        m.fs1.turbine.outlet.pressure[0].fix((P_out+1)*units.bar)

        return m
    return _make_case

def assert_willans_turbine_solution(m):
    for i in range(10):
        n_in = float(i+1) / 10 * m.fs1.turbine.willans_max_mol[0]
        m.fs1.turbine.inlet.flow_mol.fix(
            n_in
        )
        solver = SolverFactory("ipopt")
        solver.options = {"tol": 1e-3, "max_iter": 5000}
        assert degrees_of_freedom(m.fs1) == 0
        result = solver.solve(m, tee=False)
        assert result.solver.termination_condition == pyo.TerminationCondition.optimal
        assert_units_consistent(m.fs1)

def test_isentropic_turbine(build_model_with_inputs):
    m = build_model_with_inputs("isentropic")

    m.fs1.turbine.efficiency_motor.fix(1.0) 
    m.fs1.turbine.efficiency_isentropic.fix(1)

    solver = SolverFactory("ipopt")
    solver.options = {"tol": 1e-3, "max_iter": 5000}
    assert degrees_of_freedom(m.fs1) == 0
    result = solver.solve(m, tee=False)
    
    assert_units_consistent(m.fs1)

    m.fs1.turbine.report()
    
    assert result.solver.termination_condition == TerminationCondition.optimal
    
    assert m.fs1.turbine.ratioP[0].value == pytest.approx(0.2695035460992908, rel=0.001)
    assert pyo.value(m.fs1.turbine.efficiency_isentropic[0]) == pytest.approx(1, rel=0.001)
    assert m.fs1.turbine.efficiency_motor[0].value == pytest.approx(1.0, rel=0.001)
    assert m.fs1.turbine.work_electrical[0].value == pytest.approx(-16638744.441924531, rel=0.001)

def test_simple_willans_turbine(build_model_with_inputs):
    m = build_model_with_inputs("simple_willans")

    m.fs1.turbine.efficiency_motor.fix(1.0) 
    m.fs1.turbine.willans_slope.fix(190*18*units.J/units.mol)  # Willans slope 
    m.fs1.turbine.willans_intercept.fix(0.1366*1000*units.W)  # Willans intercept
    m.fs1.turbine.willans_max_mol.fix(100*15.4*units.mol / units.s)  # Willans intercept

    solver = SolverFactory("ipopt")
    solver.options = {"tol": 1e-3, "max_iter": 5000}
    assert degrees_of_freedom(m.fs1) == 0
    result = solver.solve(m, tee=False)
    
    assert_units_consistent(m.fs1)

    m.fs1.turbine.report()
    
    assert result.solver.termination_condition == TerminationCondition.optimal

    assert m.fs1.turbine.ratioP[0].value == pytest.approx(0.2695035460992908, rel=0.001)
    assert pyo.value(m.fs1.turbine.efficiency_isentropic[0]) == pytest.approx(0.5926537201367845, rel=0.001)
    assert m.fs1.turbine.efficiency_motor[0].value == pytest.approx(1.0, rel=0.001)
    assert m.fs1.turbine.work_electrical[0].value == pytest.approx(-9861013.778938433, rel=0.001)
    assert m.fs1.turbine.willans_slope[0].value == pytest.approx(3420.0, rel=0.001)
    assert m.fs1.turbine.willans_intercept[0].value == pytest.approx(136.6, rel=0.001)
    assert m.fs1.turbine.willans_max_mol[0].value == pytest.approx(1540.0, rel=0.001)

    assert_willans_turbine_solution(m)

def test_part_load_willans_turbine(build_model_with_inputs):
    
    m = build_model_with_inputs("part_load_willans")

    m.fs1.turbine.efficiency_motor.fix(1.0) 
    m.fs1.turbine.willans_max_mol.fix(217.4*15.4)  #
    m.fs1.turbine.willans_a.fix(1.5435)  # Willans slope 
    m.fs1.turbine.willans_b.fix(0.2*units.kW)  # Willans intercept
    m.fs1.turbine.willans_efficiency.fix(1 / (0.3759 + 1))  # Willans intercept

    m.fs1.turbine.report()

    assert_willans_turbine_solution(m)

def test_tsat_willans_turbine(build_model_with_inputs):
    m = build_model_with_inputs("Tsat_willans")

    m.fs1.turbine.efficiency_motor.fix(1.0) 
    m.fs1.turbine.willans_max_mol.fix(217.4*15.4)     

    solver = SolverFactory("ipopt")
    solver.options = {"tol": 1e-3, "max_iter": 5000}
    assert degrees_of_freedom(m.fs1) == 0
    result = solver.solve(m, tee=False)

    from pyomo.util.check_units import assert_units_consistent
    import pytest
    
    assert_units_consistent(m.fs1)

    m.fs1.turbine.report()
    
    assert result.solver.termination_condition == TerminationCondition.optimal
    
    assert m.fs1.turbine.ratioP[0].value == pytest.approx(0.2695035460992908, rel=0.001)
    assert pyo.value(m.fs1.turbine.efficiency_isentropic[0]) == pytest.approx(0.8000887976702635, rel=0.001)
    assert m.fs1.turbine.efficiency_motor[0].value == pytest.approx(1.0, rel=0.001)
    assert m.fs1.turbine.work_electrical[0].value == pytest.approx(-13312473.036723418, rel=0.001)
    assert m.fs1.turbine.willans_slope[0].value == pytest.approx(5724.721486921732, rel=0.001)
    assert m.fs1.turbine.willans_intercept[0].value == pytest.approx(3194420.3120209095, rel=0.001)
    assert m.fs1.turbine.willans_max_mol[0].value == pytest.approx(3347.96, rel=0.001)
    assert m.fs1.turbine.willans_a[0].value == pytest.approx(1.1916052938465973, rel=0.001)
    assert m.fs1.turbine.willans_b[0].value == pytest.approx(287807.42187938065, rel=0.001)
    assert m.fs1.turbine.willans_efficiency[0].value == pytest.approx(0.83333, rel=0.001)

    assert_willans_turbine_solution(m)

def test_bpst_willians_turbine(build_model_with_inputs):
    m = build_model_with_inputs("BPST_willans")

    m.fs1.turbine.efficiency_motor.fix(1.0) 
    m.fs1.turbine.willans_max_mol.fix(217.4*15.4) 

    solver = SolverFactory("ipopt")
    solver.options = {"tol": 1e-3, "max_iter": 5000}
    assert degrees_of_freedom(m.fs1) == 0
    result = solver.solve(m, tee=False)

    from pyomo.util.check_units import assert_units_consistent
    import pytest
    
    assert_units_consistent(m.fs1)

    m.fs1.turbine.report()
    
    assert result.solver.termination_condition == TerminationCondition.optimal
    
    assert m.fs1.turbine.ratioP[0].value == pytest.approx(0.2695035460992908, rel=0.001)
    assert pyo.value(m.fs1.turbine.efficiency_isentropic[0]) == pytest.approx(0.7640203470049349, rel=0.001)
    assert m.fs1.turbine.efficiency_motor[0].value == pytest.approx(1.0, rel=0.001)
    assert m.fs1.turbine.work_electrical[0].value == pytest.approx(-12712339.29641526, rel=0.001)
    assert m.fs1.turbine.willans_slope[0].value == pytest.approx(5511.348537597738, rel=0.001)
    assert m.fs1.turbine.willans_intercept[0].value == pytest.approx(3179303.3709413796, rel=0.001)
    assert m.fs1.turbine.willans_max_mol[0].value == pytest.approx(3347.96, rel=0.001)
    assert m.fs1.turbine.willans_a[0].value == pytest.approx(1.2284271712000001, rel=0.001)
    assert m.fs1.turbine.willans_b[0].value == pytest.approx(558672.9707596999, rel=0.001)
    assert m.fs1.turbine.willans_efficiency[0].value == pytest.approx(0.8276966055721292, rel=0.001)

    assert_willans_turbine_solution(m)

def test_ct_willians_turbine(build_model_with_inputs):
    m = build_model_with_inputs("CT_willans")

    m.fs1.turbine.inlet.enth_mol[0].fix(
        value(
            m.fs1.water.htpx(
                T=(430 + 273.15) * units.K, 
                p=6000 * units.kPa,
            )
        )
    )
    m.fs1.turbine.inlet.flow_mol[0].fix(
        2000 * units.mol / units.s
    )
    m.fs1.turbine.inlet.pressure[0].fix(
        6000 * units.kPa
    )
    m.fs1.turbine.outlet.pressure[0].fix(
        6 * units.kPa
    )

    m.fs1.turbine.efficiency_motor.fix(1.0) 
    m.fs1.turbine.willans_max_mol.fix(2000) 

    solver = SolverFactory("ipopt")
    solver.options = {"tol": 1e-3, "max_iter": 5000}
    assert degrees_of_freedom(m.fs1) == 0
    result = solver.solve(m, tee=False)
    
    assert_units_consistent(m.fs1)


import traceback
from typing import Any

from pyomo.environ import Block, value as pyo_value, Component, Var, ScalarVar, Reference, Expression
from pyomo.core.base.expression import ScalarExpression
from pyomo.core.base.constraint import ScalarConstraint
from pyomo.core.base.units_container import units
from idaes.core import FlowsheetBlock
from ahuora_builder.properties_manager import PropertiesManager
from ahuora_builder_types.unit_model_schema import SolvedPropertyValueSchema
from .units_handler import attach_unit, get_attached_unit, get_attached_unit_str, ValueWithUnits
from ahuora_builder_types import PropertiesSchema, PropertySchema
from typing import Dict
from pyomo.core.base.indexed_component import UnindexedComponent_set, IndexedComponent
from pyomo.core.base.indexed_component_slice import (
    IndexedComponent_slice,
)
import numpy as np
from ahuora_builder_types.id_types import PropertyValueId

def get_component(blk: Block, key: str):
    """
    Get a component from a block, given a key. Doesn't handle indexes.
    """
    try:
        # allow key to be split by "." to access nested properties eg. "hot_side.deltaP"
        key_split = key.split(".")
        b = blk
        for k in key_split:
            b = getattr(b, k)
        return b
    except AttributeError:
        raise ValueError(
            f"Property {key} not found in block `{blk}`. "
            f"Available properties are {[x for x in blk.component_map().keys()]}"
        )

def add_to_property_map(vars: IndexedComponent, id: PropertyValueId, fs):
    try:
        name = next(vars.values()).name # maybe we should deprectate the name?
        fs.properties_map.add(id, vars, name)
    except StopIteration:
        pass # The only reason this would happen is if there are no values in the indexed component, e.g in milk solids, there is no gas phase. In these case, we just skip adding to the property map. 
    #return c

def add_corresponding_constraint(fs: FlowsheetBlock,c, id: PropertyValueId):
    fs.properties_map.add_constraint(id, c)


def soft_cast_float(value: Any) -> float:
    """
    Softly cast a value to float, returning None if the value is None or cannot be cast.
    This is needed because not everything is indexed by time at all.
    """
    try:
        return float(value)
    except (ValueError, TypeError):
        return None

def items_by_time(s: Dict[str, Any]) -> list[tuple[str, Any]]:
    """
    Converts a dictionary of items indexed by time (as strings) to a list of tuples, and fixes the ordering.
    This is because if "11" is ordered before "2" in a dictionary of strings, but we want the results ordered by time.
    """
    return sorted(s.items(), key=lambda x: soft_cast_float(x[0]))  # sort by time index



def get_index_set_shape(component: IndexedComponent) -> tuple[int,...]:
    index_set = component.index_set()
    if index_set.dimen == 0:
        return () # I don't think we ever actually have this case (we early return if there are no items in the index set with Warning: No variables found)
    elif index_set.dimen == 1:
        return (len(index_set),)
    elif index_set.dimen > 1:
        shape = [len(s) for s in index_set.subsets()]
        return tuple(shape)


def serialize_properties_map(fs: FlowsheetBlock) -> list[SolvedPropertyValueSchema]:
    properties : list[SolvedPropertyValueSchema] = []
    # TODO: collate properties by timestep.
    properties_map: PropertiesManager = fs.properties_map
    for (uid, s) in properties_map.items():
        if uid == -1:
            # skip
            continue
        s.component

        shape = get_index_set_shape(s.component)

        values = [pyo_value(c) for c in s.component.values()]

        if shape == ():
            # not indexed at all, just return the value (not as an array)
            items = values[0]
        else:
            # Convert to an ndarray
            items = np.array(values).reshape(shape).tolist()
 
        property_dict = SolvedPropertyValueSchema(
            id=uid,
            name=s.name,  # for debugging
            value=items,
            unit=get_attached_unit_str(s.component),
        )
        if s.unknown_units:
            property_dict.unknown_units = s.unknown_units

        properties.append(property_dict)

    return properties

def slice_is_indexed(blk: IndexedComponent | IndexedComponent_slice) -> bool:
    """
    Check if a block, variable, or expression is not indexed.
    """
    if isinstance(blk, IndexedComponent_slice):
        # IndexedComponent_slice doesn't have is_indexed() method.
        # Instead, it runs is_indexed() on each of the underlying components.
        # We can assume that they are all the same, and just check the first one.
        return blk.is_indexed()[0]
    return blk.is_indexed()

def slice_index_dimen(blk: IndexedComponent | IndexedComponent_slice) -> int:
    """
    Get the dimension of the index set of a block, variable, or expression.
    """
    if isinstance(blk, IndexedComponent_slice):
        # IndexedComponent_slice doesn't have index_set() method.
        # Instead, it runs index_set() on each of the underlying components.
        # We can assume that they are all the same, and just check the first one.
        return blk.index_set()[0].dimen
    return blk.index_set().dimen


def get_sliced_version(block: Block | Var | Expression | IndexedComponent_slice) -> Block | Var | Expression | IndexedComponent_slice:
    """
    Get a sliced version of a block, variable, or expression.
    A very loose way of thinking of a sliced version is it "references all the indexes at the same time".
    
    in a scalar block, you can do scalar_block.property.
    in an indexed block, you can't do indexed_block.property, because you need to define the index you're looking at.
    However, if you want to define the indexes later, such as in a reference, you can do indexed_block[:].property.
    I.e you're creating a slice to all the indexes, and then accessing the property on that slice.

    You can't really use a sliced version directly, but you can use it to create a Reference with reference(sliced_version).
    It'll put all the indexes back together.

    The main advantage is that you can use get_sliced_version on a indexed subattribute of a slice, and it will collate the indexes together.
    """
    if not slice_is_indexed(block):
        return block
    
    block_dimen = slice_index_dimen(block)
    # we want to get a slice to all items in a block, like block[:,:]
    # To do this programmatically:
    # the ":" is represented by slice(None)
    # so for a 2D block, we want (slice(None), slice(None))
    # we can use a tuple comprehension to create this for any dimen
    block_slice = block[tuple(slice(None) for _ in range(block_dimen))]
    return block_slice


def collate_indexes(block: Block, property_key: str) -> IndexedComponent:
    """
    Returns a reference to the property with all the indexes together.
    This is because some properties have a different way of doing indexes:
    e.g 
     - properties_out[t].temperature   -> Reference[t]
     - block.heat_duty[t] -> Reference[t]
     - properties_out[t].mole_frac_comp[compound] -> Reference[t, compound]
     - block.area -> Reference[None]
    """
    block_slice = get_sliced_version(block)
    # Now we have the sliced version of the block, we can access the property using the property key.
    # As we support nested properties, e.g property_key could be "hot_side.deltaP",
    # we can use the get_component function rather than just getattr.
    block_property = get_component(block_slice, property_key)
    if block_property is None:
        raise ValueError(f"Property {property_key} not found in block {block}.")
    # Now we have the property, we need to check if it has any indexes, and get a slice to all those indexes.
    property_slice = get_sliced_version(block_property)
    #property_slice = block_property
    # Calling is_indexed() on a property_slice doesn't really make sense, it returns an array of bools. slices are always indexed though so
    # that array gets cast to True, so its' kinda okay.
    if property_slice.is_indexed(): # and not isinstance(property_slice, IndexedComponent_slice):
        # if the property is indexed, we need to return a reference to the slice.
        # This will collate all the indexes together.
        # Note that we need to special case DerivativeVar, as Reference doesn't support DerivativeVar directly. ( pyomo dae will try differentiate it again.)
        return Reference(property_slice, ctype=IndexedComponent) ##  if isinstance(block_property,DerivativeVar)  else NOTSET
    else: # this is not indexed at all lol, so we can just return the property itself. Creating a reference will add a index[None] which is unnecessary.
        return property_slice
        


def fix_var(blk: Block, var : ScalarVar | ScalarExpression, value : ValueWithUnits) -> ScalarVar | ScalarConstraint:#
    # returns: the expression that a constraint was added for, or the fixed var
    if hasattr(blk, "constrain_component"):
        component = blk.constrain_component(var, value)
    elif isinstance(var, ScalarExpression):
        component = var
    else:
        #print(var)
        var.fix(value)
        component = var
    return component

def fix_slice(var_slice: IndexedComponent | IndexedComponent_slice, values: list[ValueWithUnits]) -> list[ScalarVar | ScalarConstraint]:
    # Fix a slice of a variable to the given values.
    # the var_slice should be a Reference or other indexed component, or a scalar variable
    # values should be a list of values to fix the variable to.
    results: list[ScalarVar | ScalarConstraint] = []
    for var, value in zip(var_slice.values(), values):
        blk = var.parent_block()
        constraint = fix_var(blk, var, value)
        results.append(constraint)
    return results


def deactivate_components(components: list[ScalarVar | ScalarConstraint ]):
    # Deactivate a "PropertyValue" (which may have multiple subcomponents if it's indexed)
    for c in components:
        deactivate_component(c)

def deactivate_component(c: ScalarVar | ScalarConstraint):
    # deactivate "guess" variables: fixed for initialisation
    # and unfixed for a control constraint
    if isinstance(c, ScalarConstraint):
        c.deactivate()
    else:
        c.unfix()

def deactivate_fixed_guesses(guess_vars: list[list[ScalarVar | ScalarConstraint]]):

    for c in guess_vars:
        deactivate_components(c)


def load_initial_guess(component: Component, value: float):
    # load the initial guess into the component
    if isinstance(component, Var):
        component.value = value

def load_initial_guesses(components: IndexedComponent, values: list[float]):
    for c, v in zip(components.values(), values):
        load_initial_guess(c, v)


def fix_block(
    block: Block,
    properties_schema: PropertiesSchema,
    fs: FlowsheetBlock,
    block_ctx: "BlockContext",
) -> None:
    """
    Fix the properties of a block based on the properties schema.

    Args:
    - block: The block to fix the properties of.
    - properties_schema: The schema of the properties to fix.
    - fs: Used to store the properties in the properties map.
    """
    property_key: str
    property_info: PropertySchema
    for property_key, property_info in properties_schema.items():
        # Key is e.g "enth_mol"

        # TODO: Handle transformers
        # indexed_data = extract_indexes(
        #     property_info.data,
        #     property_info.unit,
        #     transformers,
        # )
        property_reference = collate_indexes(block, property_key)

        for property_value in property_info.data:
            discrete_indexes = property_value.discrete_indexes or []

            pv_id = property_value.id
            num_discrete_indexes = len(discrete_indexes)
            num_property_indexes = property_reference.index_set().dimen if property_reference.is_indexed() else 0
            num_continuous_indexes = num_property_indexes - num_discrete_indexes # This is the dimension of the property_value.value ndarray.

            # We have the convention that all the continuous indexes come first in property_reference, and then the discrete indexes.
            # This is what idaes normally does.
            
            
            # We need to get a slice to the current set of discrete indexes.
            if len(discrete_indexes) == 0:
                property_slice = property_reference # no indexes to worry about.
            else:
                property_slice = property_reference[tuple(
                    list(slice(None) for _ in range(num_continuous_indexes)) + discrete_indexes
                )]
            # Now property_slice is only indexed by the continuous indexes.
            
            # get a reference to the variable/expression. This will also add an index [None] if it is not indexed at all, i.e no continuous indexes.
            variable_references = Reference(property_slice, ctype=IndexedComponent) #Ctype=IndexedComponent avoids problems with DerivativeVar
            
            add_to_property_map(variable_references, pv_id, fs)
            
            # Because both pyomo and numpy flatten arrays with the last index changing fastest, we can just flatten both the index set and the values, and then iterate through them together.
            variable_indexes = list(variable_references.index_set())

            if (len(variable_indexes) == 0):
                print("Warning: No variables found for {property_key} with indexes {discrete_indexes}. This may be expected in the milk property package, which doesn't have a gas phase.")
                continue
            
            variable_transformed = variable_references
            

            if property_value.value is not None:
                variable_values = np.array([property_value.value]).flatten() # we put the value in an array to handle the case where it is a scalar.
                variable_values_with_units = [attach_unit(v, property_info.unit) for v in variable_values]
                variable_values_converted = [
                    units.convert(v, get_attached_unit(var)) for v, var in zip(variable_values_with_units, variable_references.values())
                ]

                #value = units.convert(property_value.value, get_attached_unit(var))
                if property_value.constraint is not None:
                    # add the constraint to the list of constraints to be added
                    # at the flowsheet level. The var value should also
                    # be a guess to maintain the degrees of freedom.
                    expr = property_value.constraint
                    fs.constraint_exprs.append((variable_transformed, expr, pv_id))
                    if property_value.controlled:
                        # this is a set point. To maintain degrees of freedom,
                        # use the manipulated variable as a guess during initialisation.
                        # this is a weird case because we might have guesses for both the
                        # manipulated variable and this variable, and we want to end up
                        # using the formula/constraint. If we were to try use this guess
                        # by eliminating the manipulated variable guess, we would run into
                        # degrees of freedom issues at the flowsheet level if elimination
                        # failed (since it then adds another flowsheet level constraint).
                        load_initial_guesses(variable_references, variable_values_converted)
                    else:
                        # not a set point, use this guess during initialization.
                        component = fix_slice(variable_references, variable_values_converted)
                        fs.guess_vars.append(component)
                elif property_value.controlled is not None:
                    # this value is a controlling variable, so don't fix it now
                    # it should be fixed after initialisation
                    block_ctx.add_controlled_var(variable_references, variable_values_converted, pv_id, property_value.controlled)
                elif property_value.guess:
                    block_ctx.add_guess_var( variable_references, variable_values_converted, pv_id)
                else:
                    components = fix_slice(variable_references, variable_values_converted)
                    # add_corresponding_constraint used to take the constraint returned by constrain_component if you are constraining an expression. TODO: Do we need to add this back in?
                    # from c = fix_var()
                    add_corresponding_constraint(fs, components, pv_id)



from ahuora_builder.methods.BlockContext import BlockContext


#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Ben Lincolns heat exchanger model using effectiveness

"""

# Import Pyomo libraries
from pyomo.environ import (
    Block,
    check_optimal_termination,
    Constraint,
    Expression,
    Param,
    PositiveReals,
    Reference,
    units as pyunits,
    Var,
)
from pyomo.common.config import Bool, ConfigBlock, ConfigValue, In

# Import IDAES cores
from idaes.core import (
    ControlVolume0DBlock,
    declare_process_block_class,
    MaterialBalanceType,
    EnergyBalanceType,
    MomentumBalanceType,
    UnitModelBlockData,
    useDefault,
)
from idaes.models.unit_models.heat_exchanger import hx_process_config, add_hx_references
from idaes.core.util.config import is_physical_parameter_block
from idaes.core.util.tables import create_stream_table_dataframe
from idaes.core.util.math import smooth_min, smooth_max
from idaes.core.solvers import get_solver
from idaes.core.util.exceptions import InitializationError
import idaes.logger as idaeslog
from idaes.core.initialization import SingleControlVolumeUnitInitializer
from idaes.core.util.model_statistics import degrees_of_freedom
from .inverted import add_inverted, initialise_inverted

__author__ = "Paul Akula, Andrew Lee, Ben Lincoln"


# Set up logger
_log = idaeslog.getLogger(__name__)


class HXEFFInitializer(SingleControlVolumeUnitInitializer):
    """
    Initializer for NTU Heat Exchanger units.

    """

    def initialization_routine(
        self,
        model: Block,
        plugin_initializer_args: dict = None,
        copy_inlet_state: bool = False,
        duty=1000 * pyunits.W,
    ):
        """
        Common initialization routine for NTU Heat Exchangers.

        This routine starts by initializing the hot and cold side properties. Next, the heat
        transfer between the two sides is fixed to an initial guess for the heat duty (provided by the duty
        argument), the associated constraint deactivated, and the model is then solved. Finally, the heat
        duty is unfixed and the heat transfer constraint reactivated followed by a final solve of the model.

        Args:
            model: Pyomo Block to be initialized
            plugin_initializer_args: dict-of-dicts containing arguments to be passed to plug-in Initializers.
                Keys should be submodel components.
            copy_inlet_state: bool (default=False). Whether to copy inlet state to other states or not
                (0-D control volumes only). Copying will generally be faster, but inlet states may not contain
                all properties required elsewhere.
            duty: initial guess for heat duty to assist with initialization. Can be a Pyomo expression with units.

        Returns:
            Pyomo solver results object
        """
        return super(SingleControlVolumeUnitInitializer, self).initialization_routine(
            model=model,
            plugin_initializer_args=plugin_initializer_args,
            copy_inlet_state=copy_inlet_state,
            duty=duty,
        )

    def initialize_main_model(
        self,
        model: Block,
        copy_inlet_state: bool = False,
        duty=1000 * pyunits.W,
    ):
        """
        Initialization routine for main NTU HX models.

        Args:
            model: Pyomo Block to be initialized.
            copy_inlet_state: bool (default=False). Whether to copy inlet state to other states or not
                (0-D control volumes only). Copying will generally be faster, but inlet states may not contain
                all properties required elsewhere.
            duty: initial guess for heat duty to assist with initialization. Can be a Pyomo expression with units.

        Returns:
            Pyomo solver results object.

        """
        initialise_inverted(model.hot_side,"deltaP")
        initialise_inverted(model.cold_side,"deltaP")
        # TODO: Aside from one differences in constraint names, this is
        # identical to the Initializer for the 0D HX unit.
        # Set solver options
        init_log = idaeslog.getInitLogger(
            model.name, self.get_output_level(), tag="unit"
        )
        solve_log = idaeslog.getSolveLogger(
            model.name, self.get_output_level(), tag="unit"
        )

        # Create solver
        solver = self._get_solver()

        self.initialize_control_volume(model.hot_side, copy_inlet_state)
        init_log.info_high("Initialization Step 1a (hot side) Complete.")

        self.initialize_control_volume(model.cold_side, copy_inlet_state)
        init_log.info_high("Initialization Step 1b (cold side) Complete.")

        # ---------------------------------------------------------------------
        # Solve unit without heat transfer equation
        model.energy_balance_constraint.deactivate()

        model.cold_side.heat.fix(duty)
        for i in model.hot_side.heat:
            model.hot_side.heat[i].set_value(-duty)

        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = solver.solve(model, tee=slc.tee)

        init_log.info_high("Initialization Step 2 {}.".format(idaeslog.condition(res)))

        model.cold_side.heat.unfix()
        model.energy_balance_constraint.activate()
        # ---------------------------------------------------------------------
        # Solve unit
        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = solver.solve(model, tee=slc.tee)
        init_log.info("Initialization Completed, {}".format(idaeslog.condition(res)))

        return res


@declare_process_block_class("HeatExchangerEffectiveness")
class HeatExchangerEFFData(UnitModelBlockData):
    """Heat Exchanger Unit Model using NTU method."""

    default_initializer = HXEFFInitializer

    CONFIG = UnitModelBlockData.CONFIG(implicit=True)

    # Configuration template for fluid specific  arguments
    _SideCONFIG = ConfigBlock()
    _SideCONFIG.declare(
        "has_phase_equilibrium",
        ConfigValue(
            default=False,
            domain=Bool,
            description="Phase equilibrium construction flag",
            doc="""Indicates whether terms for phase equilibrium should be
constructed, **default** = False.
**Valid values:** {
**True** - include phase equilibrium terms
**False** - exclude phase equilibrium terms.}""",
        ),
    )
    _SideCONFIG.declare(
        "material_balance_type",
        ConfigValue(
            default=MaterialBalanceType.useDefault,
            domain=In(MaterialBalanceType),
            description="Material balance construction flag",
            doc="""Indicates what type of mass balance should be constructed,
**default** - MaterialBalanceType.useDefault.
**Valid values:** {
**MaterialBalanceType.useDefault - refer to property package for default
balance type
**MaterialBalanceType.none** - exclude material balances,
**MaterialBalanceType.componentPhase** - use phase component balances,
**MaterialBalanceType.componentTotal** - use total component balances,
**MaterialBalanceType.elementTotal** - use total element balances,
**MaterialBalanceType.total** - use total material balance.}""",
        ),
    )
    _SideCONFIG.declare(
        "energy_balance_type",
        ConfigValue(
            default=EnergyBalanceType.useDefault,
            domain=In(EnergyBalanceType),
            description="Energy balance construction flag",
            doc="""Indicates what type of energy balance should be constructed,
**default** - EnergyBalanceType.useDefault.
**Valid values:** {
**EnergyBalanceType.useDefault - refer to property package for default
balance type
**EnergyBalanceType.none** - exclude energy balances,
**EnergyBalanceType.enthalpyTotal** - single enthalpy balance for material,
**EnergyBalanceType.enthalpyPhase** - enthalpy balances for each phase,
**EnergyBalanceType.energyTotal** - single energy balance for material,
**EnergyBalanceType.energyPhase** - energy balances for each phase.}""",
        ),
    )
    _SideCONFIG.declare(
        "momentum_balance_type",
        ConfigValue(
            default=MomentumBalanceType.pressureTotal,
            domain=In(MomentumBalanceType),
            description="Momentum balance construction flag",
            doc="""Indicates what type of momentum balance should be constructed,
**default** - MomentumBalanceType.pressureTotal.
**Valid values:** {
**MomentumBalanceType.none** - exclude momentum balances,
**MomentumBalanceType.pressureTotal** - single pressure balance for material,
**MomentumBalanceType.pressurePhase** - pressure balances for each phase,
**MomentumBalanceType.momentumTotal** - single momentum balance for material,
**MomentumBalanceType.momentumPhase** - momentum balances for each phase.}""",
        ),
    )
    _SideCONFIG.declare(
        "has_pressure_change",
        ConfigValue(
            default=False,
            domain=Bool,
            description="Pressure change term construction flag",
            doc="""Indicates whether terms for pressure change should be
constructed,
**default** - False.
**Valid values:** {
**True** - include pressure change terms,
**False** - exclude pressure change terms.}""",
        ),
    )
    _SideCONFIG.declare(
        "property_package",
        ConfigValue(
            default=useDefault,
            domain=is_physical_parameter_block,
            description="Property package to use ",
            doc="""Property parameter object used to define property calculations
        **default** - useDefault.
        **Valid values:** {
        **useDefault** - use default package from parent model or flowsheet,
        **PhysicalParameterObject** - a PhysicalParameterBlock object.}""",
        ),
    )
    _SideCONFIG.declare(
        "property_package_args",
        ConfigBlock(
            implicit=True,
            description="Arguments to use for constructing property package",
            doc="""A ConfigBlock with arguments to be passed to
        property block(s) and used when constructing these,
        **default** - None.
        **Valid values:** {
        see property package for documentation.}""",
        ),
    )

    # Create individual config blocks for hot and cold sides
    CONFIG.declare("hot_side", _SideCONFIG(doc="Hot fluid config arguments"))
    CONFIG.declare("cold_side", _SideCONFIG(doc="Cold fluid config arguments"))
    CONFIG.declare(
        "hot_side_name",
        ConfigValue(
            default=None,
            domain=str,
            doc="Hot side name, sets control volume and inlet and outlet names",
        ),
    )
    CONFIG.declare(
        "cold_side_name",
        ConfigValue(
            default=None,
            domain=str,
            doc="Cold side name, sets control volume and inlet and outlet names",
        ),
    )

    def build(self):
        # Call UnitModel.build to setup model
        super().build()
        hx_process_config(self)

        # ---------------------------------------------------------------------
        # Build hot-side control volume
        self.hot_side = ControlVolume0DBlock(
            dynamic=self.config.dynamic,
            has_holdup=self.config.has_holdup,
            property_package=self.config.hot_side.property_package,
            property_package_args=self.config.hot_side.property_package_args,
        )

        # TODO : Add support for phase equilibrium?
        self.hot_side.add_state_blocks(has_phase_equilibrium=self.config.hot_side.has_phase_equilibrium)

        self.hot_side.add_material_balances(
            balance_type=self.config.hot_side.material_balance_type,
            has_phase_equilibrium=self.config.hot_side.has_phase_equilibrium,
        )

        self.hot_side.add_energy_balances(
            balance_type=self.config.hot_side.energy_balance_type,
            has_heat_transfer=True,
        )

        self.hot_side.add_momentum_balances(
            balance_type=self.config.hot_side.momentum_balance_type,
            has_pressure_change=self.config.hot_side.has_pressure_change,
        )

        # ---------------------------------------------------------------------
        # Build cold-side control volume
        self.cold_side = ControlVolume0DBlock(
            dynamic=self.config.dynamic,
            has_holdup=self.config.has_holdup,
            property_package=self.config.cold_side.property_package,
            property_package_args=self.config.cold_side.property_package_args,
        )

        self.cold_side.add_state_blocks(has_phase_equilibrium=self.config.cold_side.has_phase_equilibrium)

        self.cold_side.add_material_balances(
            balance_type=self.config.cold_side.material_balance_type,
            has_phase_equilibrium=self.config.cold_side.has_phase_equilibrium,
        )

        self.cold_side.add_energy_balances(
            balance_type=self.config.cold_side.energy_balance_type,
            has_heat_transfer=True,
        )

        self.cold_side.add_momentum_balances(
            balance_type=self.config.cold_side.momentum_balance_type,
            has_pressure_change=self.config.cold_side.has_pressure_change,
        )

        # ---------------------------------------------------------------------
        # Add Ports to control volumes
        self.add_inlet_port(
            name="hot_side_inlet", block=self.hot_side, doc="Hot side inlet port"
        )
        self.add_outlet_port(
            name="hot_side_outlet", block=self.hot_side, doc="Hot side outlet port"
        )

        self.add_inlet_port(
            name="cold_side_inlet", block=self.cold_side, doc="Cold side inlet port"
        )
        self.add_outlet_port(
            name="cold_side_outlet", block=self.cold_side, doc="Cold side outlet port"
        )

        # ---------------------------------------------------------------------
        # Add unit level References
        # Set references to balance terms at unit level
        self.heat_duty = Reference(self.cold_side.heat[:])

        # Add references to the user provided aliases (if applicable).
        add_hx_references(self)

        # ---------------------------------------------------------------------
        # Add performance equations
        # All units of measurement will be based on hot side
        hunits = self.hot_side.config.property_package.get_metadata().get_derived_units

        # Overall energy balance
        def rule_energy_balance(blk, t):
            return blk.hot_side.heat[t] == -pyunits.convert(
                blk.cold_side.heat[t], to_units=hunits("power")
            )

        self.energy_balance_constraint = Constraint(
            self.flowsheet().time, rule=rule_energy_balance
        )

        # Add e-NTU variables
        self.effectiveness = Var(
            self.flowsheet().time,
            initialize=1,
            units=pyunits.dimensionless,
            domain=PositiveReals,
            doc="Effectiveness factor",
        )

        # Minimum heat capacitance ratio for e-NTU method
        self.eps_cmin = Param(
            initialize=1e-3,
            mutable=True,
            units=hunits("power") / hunits("temperature"),
            doc="Epsilon parameter for smooth Cmin and Cmax",
        )
        
###########

        # Todo: Will not work for a pure component system in the two phase region

###########

        #Build hotside State Block
        tmp_dict = dict(**self.config.hot_side.property_package_args)
        tmp_dict["has_phase_equilibrium"] = self.config.hot_side.has_phase_equilibrium
        tmp_dict["defined_state"] = False

        self.properties_hotside = self.config.hot_side.property_package.build_state_block(
            self.flowsheet().time, doc="Hot side properties at outlet under Qmax", **tmp_dict
        )
        #Connect Properties to Hot Side 
        @self.Constraint(
                self.flowsheet().time, doc="Hot Side Temperature at Qmax = Inlet Cold Temp")
        def constraint_hotside_temp(self, t):
            return self.properties_hotside[t].temperature == self.cold_side.properties_in[t].temperature
        @self.Constraint(
                self.flowsheet().time, doc="Hot Side Pressure at Qmax = Inlet Hot Pressure")
        def constraint_hotside_pres(self, t):
            return self.properties_hotside[t].pressure == self.hot_side.properties_in[t].pressure
        @self.Constraint(
                self.flowsheet().time, doc="Hot Side Flowrate at Qmax = Inlet Hot Flowrate")
        def constraint_hotside_flow(self, t):
            return self.properties_hotside[t].flow_mol == self.hot_side.properties_in[t].flow_mol
        
        #Build Coldside State Block
        tmp_dict = dict(**self.config.cold_side.property_package_args)
        tmp_dict["has_phase_equilibrium"] = self.config.cold_side.has_phase_equilibrium
        tmp_dict["defined_state"] = False

        self.properties_coldside = self.config.cold_side.property_package.build_state_block(
            self.flowsheet().time, doc="Cold side properties at outlet under Qmax", **tmp_dict
        )

########

        # Todo: Will not work for a pure component system in the two phase region

###########

        #Connect Properties to Cold Side
        @self.Constraint(
                self.flowsheet().time, doc="Cold Side Temperature at Qmax = Inlet Hot Temp")
        def constraint_coldside_temp(self, t):
            return self.properties_coldside[t].temperature == self.hot_side.properties_in[t].temperature
        @self.Constraint(
                self.flowsheet().time, doc="Cold Side Pressure at Qmax = Inlet Cold Pressure")
        def constraint_coldside_pres(self, t):
            return self.properties_coldside[t].pressure == self.cold_side.properties_in[t].pressure
        @self.Constraint(
                self.flowsheet().time, doc="Cold Side Flowrate at Qmax = Inlet Cold Flowrate")
        def constraint_coldside_flow(self, t):
            return self.properties_coldside[t].flow_mol == self.cold_side.properties_in[t].flow_mol
        
        # Delta h hot at Qmax
        def rule_deltah_hot_qmax(blk, t):
            return blk.hot_side.properties_in[t].enth_mol - blk.properties_hotside[t].enth_mol

        self.delta_h_hot_qmax = Expression(
            self.flowsheet().time, rule=rule_deltah_hot_qmax, doc="Delta h Hot Side at Qmax"
        )
        # Delta h cold at Qmax
        def rule_deltah_cold_qmax(blk, t):
            return blk.properties_coldside[t].enth_mol - blk.cold_side.properties_in[t].enth_mol
        self.delta_h_cold_qmax = Expression(
            self.flowsheet().time, rule=rule_deltah_cold_qmax, doc="Delta h Cold Side at Qmax"
        )

        # TODO : Support both mass and mole based flows
        # Minimum heat transfer rate
        def rule_Hmin(blk, t):
            Hhot = pyunits.convert(
                blk.hot_side.properties_in[t].flow_mol
                * blk.delta_h_hot_qmax[t],
                to_units=hunits("power"),
            )
            Hcold = pyunits.convert(
                blk.cold_side.properties_in[t].flow_mol
                * blk.delta_h_cold_qmax[t],
                to_units=hunits("power"),
            )
            return smooth_min(Hhot, Hcold, eps=blk.eps_cmin)

        self.Qmax = Expression(
            self.flowsheet().time, rule=rule_Hmin, doc="Max heat transfer rate"
        )
        #Max heat transfer rate
        def rule_Hmax(blk, t):
            Hhot = pyunits.convert(
                blk.hot_side.properties_in[t].flow_mol
                * blk.delta_h_hot_qmax[t],
                to_units=hunits("power"),
            )
            Hcold = pyunits.convert(
                blk.cold_side.properties_in[t].flow_mol
                * blk.delta_h_cold_qmax[t],
                to_units=hunits("power"),
            )
            return smooth_max(Hhot, Hcold, eps=blk.eps_cmin)
        self.Hmax_theory = Expression(
            self.flowsheet().time,
              rule=rule_Hmax, doc="Maximum heat transfer rate"
        )

# Add effectiveness relation
        def rule_effectiveness(blk, t):
            return blk.hot_side.heat[t] == -(
                blk.effectiveness[t]
                * blk.Qmax[t]
            )

        self.heat_duty_constraint = Constraint(self.flowsheet().time, rule=rule_effectiveness)

        add_inverted(self.hot_side,"deltaP")
        add_inverted(self.cold_side,"deltaP")
    # TODO : Add scaling methods

    def initialize_build(
        self,
        hot_side_state_args=None,
        cold_side_state_args=None,
        outlvl=idaeslog.NOTSET,
        solver=None,
        optarg=None,
        duty=None,
    ):
        """
        Heat exchanger initialization method.

        Args:
            hot_side_state_args : a dict of arguments to be passed to the
                property initialization for the hot side (see documentation of
                the specific property package) (default = None).
            cold_side_state_args : a dict of arguments to be passed to the
                property initialization for the cold side (see documentation of
                the specific property package) (default = None).
            outlvl : sets output level of initialization routine
            optarg : solver options dictionary object (default=None, use
                     default solver options)
            solver : str indicating which solver to use during
                     initialization (default = None, use default solver)
            duty : an initial guess for the amount of heat transferred. This
                should be a tuple in the form (value, units), (default
                = (1000 J/s))

        Returns:
            None

        """
        initialise_inverted(self.hot_side,"deltaP")
        initialise_inverted(self.cold_side,"deltaP")
        # Set solver options
        init_log = idaeslog.getInitLogger(self.name, outlvl, tag="unit")
        solve_log = idaeslog.getSolveLogger(self.name, outlvl, tag="unit")

        hot_side = self.hot_side
        cold_side = self.cold_side
        hot_side_max = self.properties_hotside
        cold_side_max = self.properties_coldside
        

        # Create solver
        opt = get_solver(solver, optarg)

        flags1 = hot_side.initialize(
            outlvl=outlvl, optarg=optarg, solver=solver, state_args=hot_side_state_args
        )

        init_log.info_high("Initialization Step 1a (hot side) Complete.")

        flags2 = cold_side.initialize(
            outlvl=outlvl, optarg=optarg, solver=solver, state_args=cold_side_state_args
        )

        init_log.info_high("Initialization Step 1b (cold side) Complete.")

        # cold_side_max[0].pressure.set_value(cold_side.properties_in[0].pressure.value)
        # hot_side_max[0].pressure.set_value(hot_side.properties_in[0].pressure.value)

        # cold_side_max[0].flow_mol.set_value(cold_side.properties_in[0].flow_mol.value)
        # hot_side_max[0].flow_mol.set_value(hot_side.properties_in[0].flow_mol.value)

        # cold_side_max[0].enth_mol.set_value(cold_side.properties_in[0].enth_mol.value)
        # hot_side_max[0].enth_mol.set_value(hot_side.properties_in[0].enth_mol.value)

        flags3 = hot_side_max.initialize(
            outlvl=outlvl
        )
        init_log.info_high("Initialization Step 1c (hot side max) Complete.")

        flags4 = cold_side_max.initialize(
            outlvl=outlvl
        )
        init_log.info_high("Initialization Step 1d (cold side max) Complete.")

        hot_side_max[0].display()
        cold_side_max[0].display()


        #----------------------------------------------------------------------
       
        self.energy_balance_constraint.deactivate()
        self.heat_duty_constraint.deactivate()
        self.hot_side.enthalpy_balances.deactivate()
        self.cold_side.enthalpy_balances.deactivate()
        self.cold_side.material_balances.deactivate()
        self.hot_side.material_balances.deactivate()
        self.hot_side.properties_out.deactivate()
        self.cold_side.properties_out.deactivate()

        print("DOF:", degrees_of_freedom(self))
        from idaes.core.util import DiagnosticsToolbox
        dt = DiagnosticsToolbox(self)
        dt.display_underconstrained_set()
        dt.display_overconstrained_set()
        dt.report_structural_issues()
        dt.report_numerical_issues()
        dt.display_constraints_with_large_residuals()
        #dt.compute_infeasibility_explanation()
        dt.display_variables_at_or_outside_bounds()

        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = opt.solve(self, tee=slc.tee)
            

        init_log.info_high("Initialization Step 1.5 {}.".format(idaeslog.condition(res)))

        # ---------------------------------------------------------------------
        # Solve unit without heat transfer equation
        self.heat_duty_constraint.activate()
        self.heat_duty.unfix()
        self.hot_side.enthalpy_balances.activate()
        self.cold_side.enthalpy_balances.activate()
        self.cold_side.material_balances.activate()
        self.hot_side.material_balances.activate()
        self.hot_side.properties_out.activate()
        self.cold_side.properties_out.activate()
        
        # self.Qmax[0].set_value(1000)

        # Get side 1 and side 2 heat units, and convert duty as needed
        s1_units = hot_side.heat.get_units()
        s2_units = cold_side.heat.get_units()

        if duty is None:
            # Assume 1000 J/s and check for unitless properties
            if s1_units is None and s2_units is None:
                # Backwards compatibility for unitless properties
                s1_duty = -1000
                s2_duty = 1000
            else:
                s1_duty = pyunits.convert_value(
                    -1000, from_units=pyunits.W, to_units=s1_units
                )
                s2_duty = pyunits.convert_value(
                    1000, from_units=pyunits.W, to_units=s2_units
                )
        else:
            # Duty provided with explicit units
            s1_duty = -pyunits.convert_value(
                duty[0], from_units=duty[1], to_units=s1_units
            )
            s2_duty = pyunits.convert_value(
                duty[0], from_units=duty[1], to_units=s2_units
            )

        cold_side.heat.fix(s2_duty)
        for i in hot_side.heat:
            hot_side.heat[i].value = s1_duty

        print("DOF:", degrees_of_freedom(self))

        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = opt.solve(self, tee=slc.tee)

        init_log.info_high("Initialization Step 2 {}.".format(idaeslog.condition(res)))

        cold_side.heat.unfix()
        self.energy_balance_constraint.activate()

        # ---------------------------------------------------------------------
        # Solve unit
        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = opt.solve(self, tee=slc.tee)
        init_log.info_high("Initialization Step 3 {}.".format(idaeslog.condition(res)))

        # ---------------------------------------------------------------------
        # Release Inlet state
        hot_side.release_state(flags1, outlvl=outlvl)
        cold_side.release_state(flags2, outlvl=outlvl)

        init_log.info("Initialization Completed, {}".format(idaeslog.condition(res)))

        if not check_optimal_termination(res):
            raise InitializationError(
                f"{self.name} failed to initialize successfully. Please check "
                f"the output logs for more information."
            )

        return res

    def _get_stream_table_contents(self, time_point=0):
        return create_stream_table_dataframe(
            {
                "Hot Inlet": self.hot_side_inlet,
                "Hot Outlet": self.hot_side_outlet,
                "Cold Inlet": self.cold_side_inlet,
                "Cold Outlet": self.cold_side_outlet,
            },
            time_point=time_point,
        )


from ahuora_builder.methods.adapter import add_corresponding_constraint, fix_var,fix_slice, load_initial_guess


from idaes.core import FlowsheetBlock
from pyomo.core.base.constraint import Constraint
from pyomo.environ import Block, Component, Reference
from ahuora_builder_types.id_types import PropertyValueId
from pyomo.core.base.indexed_component_slice import (
    IndexedComponent_slice,
    _IndexedComponent_slice_iter,
)
from pyomo.core.base.indexed_component import UnindexedComponent_set, IndexedComponent


class BlockContext:
    """
    Where possible, we want to fix variables at the block level (ie. unit model, state block)
    rather than at the flowsheet level. This is because it is easier to solve a smaller model
    during initialization, rather than dumping complexity on the solver when solving the entire
    flowsheet.

    Each controlling variable in the model is accompanied by a guess variable. Normally, the
    guess variable is fixed during initialization, and unfixed after, while the controlling
    variable (set point) is a constraint at the flowsheet level. However, if both the guess and
    controlling variable are on the same block, we can avoid this and fix the controlling variable
    directly at the block level.

    This class provides a context to store controlling variables and guess variables while fixing
    within a block. We can then apply a simple heuristic to eliminate as many pairs of guess and
    controlling variables as possible.
    """

    def __init__(self, flowsheet: FlowsheetBlock):
        """
        - blk: Pyomo block the Var is on (can add constraints to this block)
        - var: Pyomo Var to fix/constrain
        - value: value to fix/constrain the Var to
        - id: id of the property, to store the created Constraint in the properties map
        """
        # property id: ( var_reference, values)
        self._guess_vars: dict[PropertyValueId, tuple[ IndexedComponent | IndexedComponent_slice, list[float]]] = {}
        # property id: ( var_reference, values, guess_id)
        self._controlled_vars: dict[PropertyValueId, tuple[IndexedComponent | IndexedComponent_slice, list[float], PropertyValueId]] = {}
        self._flowsheet = flowsheet

    def add_guess_var(self, var_references : IndexedComponent | IndexedComponent_slice, values : list[float], propertyvalue_id : PropertyValueId):
        self._guess_vars[propertyvalue_id] = ( var_references, values)

    def add_controlled_var(self, var_references: IndexedComponent | IndexedComponent_slice, values: list[float], propertyvalue_id: PropertyValueId, guess_propertyvalue_id: PropertyValueId):
        self._controlled_vars[propertyvalue_id] = (var_references, values, guess_propertyvalue_id)

    def apply_elimination(self):
        """
        Try to eliminate as many guess vars/flowsheet-level constraints as possible.
        Fix the remaining guess vars or add the remaining controlled vars as constraints.
        """
        # TODO: Update apply_elimination with the lists of values now.
        fs = self._flowsheet
        for id, (var_refs, values, guess_id) in self._controlled_vars.items():
            # see if we can eliminate this controlled var
            if guess_id in self._guess_vars:
                # fix the controlled var
                c = fix_slice(var_refs, values)
                add_corresponding_constraint(fs, c, id)
                # load the initial guess for the guess var
                var_refs, values = self._guess_vars[guess_id]
                load_initial_guess(var_refs, values)
                # eliminate the guess var
                del self._guess_vars[guess_id]
            else:
                # add the control as a flowsheet-level constraint
                # As the values are flattended into a list, we also need to flatten the index set into a list.
                var_refs_list = list(var_refs.values()) # returns a list of VarData or ExpressionData objects
                def constraint_rule(blk,idx):
                    return var_refs_list[idx] == values[idx]
                c = Constraint(range(len(var_refs_list)), rule=constraint_rule)
                name = f"control_constraint_{id}" # Maybe we could use the var name or something here? but it's a bit harder with indexed constraints. remember it has to be unique!
                self._flowsheet.add_component(name, c)
                add_corresponding_constraint(fs, c, id)

        # fix the remaining guess vars
        for id, (var_refs, values) in self._guess_vars.items():
            c = fix_slice(var_refs, values)
            self._flowsheet.guess_vars.append(c)


from idaes.models.unit_models.separator import SeparatorData, SplittingType

from functools import partial
from pandas import DataFrame

from pyomo.environ import (
    Block,
    check_optimal_termination,
    Constraint,
    Param,
    Reals,
    Reference,
    Set,
    Var,
    value,
)
from pyomo.network import Port
from pyomo.common.config import ConfigBlock, ConfigValue, In, ListOf, Bool

from idaes.core import (
    declare_process_block_class,
    UnitModelBlockData,
    useDefault,
    MaterialBalanceType,
    MomentumBalanceType,
    MaterialFlowBasis,
    VarLikeExpression,
)
from idaes.core.util.config import (
    is_physical_parameter_block,
    is_state_block,
)
from idaes.core.util.exceptions import (
    BurntToast,
    ConfigurationError,
    PropertyNotSupportedError,
    InitializationError,
)
from idaes.core.solvers import get_solver
from idaes.core.util.tables import create_stream_table_dataframe
from idaes.core.util.model_statistics import degrees_of_freedom
import idaes.logger as idaeslog
import idaes.core.util.scaling as iscale
from idaes.core.util.units_of_measurement import report_quantity
from idaes.core.initialization import ModularInitializerBase


# This only changes a couple of lines in the original SeparatorData class, to not fix state variables by default.
# The state block initialisation already does this if needed, so we can just set their value.
# This is because if the state block has extra constraints, such as for flow_mass, then fixing flow_mol will over-define the system.
# It might be worth making this a pr to idaes.

@declare_process_block_class("CustomSeparator")
class CustomSeparatorData(SeparatorData):

    def initialize_build(
        blk, outlvl=idaeslog.NOTSET, optarg=None, solver=None, hold_state=False
    ):
        """
        Initialization routine for separator

        Keyword Arguments:
            outlvl : sets output level of initialization routine
            optarg : solver options dictionary object (default=None, use
                     default solver options)
            solver : str indicating which solver to use during
                     initialization (default = None, use default solver)
            hold_state : flag indicating whether the initialization routine
                     should unfix any state variables fixed during
                     initialization, **default** - False. **Valid values:**
                     **True** - states variables are not unfixed, and a dict of
                     returned containing flags for which states were fixed
                     during initialization, **False** - state variables are
                     unfixed after initialization by calling the release_state
                     method.

        Returns:
            If hold_states is True, returns a dict containing flags for which
            states were fixed during initialization.
        """
        init_log = idaeslog.getInitLogger(blk.name, outlvl, tag="unit")
        solve_log = idaeslog.getSolveLogger(blk.name, outlvl, tag="unit")

        # Create solver
        opt = get_solver(solver, optarg)

        # Initialize mixed state block
        if blk.config.mixed_state_block is not None:
            mblock = blk.config.mixed_state_block
        else:
            mblock = blk.mixed_state
        flags = mblock.initialize(
            outlvl=outlvl,
            optarg=optarg,
            solver=solver,
            hold_state=True,
        )

        # Solve for split fractions only
        component_status = {}
        for c in blk.component_objects((Block, Constraint)):
            for i in c:
                if not c[i].local_name == "sum_split_frac":
                    # Record current status of components to restore later
                    component_status[c[i]] = c[i].active
                    c[i].deactivate()

        if degrees_of_freedom(blk) != 0:
            with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
                res = opt.solve(blk, tee=slc.tee)
                init_log.info(
                    "Initialization Step 1 Complete: {}".format(idaeslog.condition(res))
                )

        for c, s in component_status.items():
            if s:
                c.activate()

        if blk.config.ideal_separation:
            # If using ideal splitting, initialization should be complete
            return flags

        # Initialize outlet StateBlocks
        outlet_list = blk.create_outlet_list()

        # Premises for initializing outlet states:
        # 1. Intensive states remain unchanged - this is either a valid premise
        # or the actual state is impossible to calculate without solving the
        # full separator model.
        # 2. Extensive states are use split fractions if index matches, or
        # average of split fractions for outlet otherwise
        for o in outlet_list:
            # Get corresponding outlet StateBlock
            o_block = getattr(blk, o + "_state")

            # Create dict to store fixed status of state variables
            o_flags = {}
            for t in blk.flowsheet().time:

                # Calculate values for state variables
                s_vars = o_block[t].define_state_vars()
                for v in s_vars:
                    for k in s_vars[v]:
                        # Record whether variable was fixed or not
                        o_flags[t, v, k] = s_vars[v][k].fixed

                        # If fixed, use current value
                        # otherwise calculate guess from mixed state and fix
                        if not s_vars[v][k].fixed:
                            m_var = getattr(mblock[t], s_vars[v].local_name)
                            if "flow" in v:
                                # If a "flow" variable, is extensive
                                # Apply split fraction
                                if blk.config.split_basis == SplittingType.totalFlow:
                                    # All flows split by outlet
                                    s_vars[v][k].set_value(
                                        value(m_var[k] * blk.split_fraction[(t, o)])
                                    )
                                elif "_phase_comp" in v:
                                    # Need to match indices, but use split frac
                                    if (
                                        blk.config.split_basis
                                        == SplittingType.phaseComponentFlow
                                    ):
                                        s_vars[v][k].set_value(
                                            value(
                                                m_var[k]
                                                * blk.split_fraction[(t, o) + (k,)]
                                            )
                                        )
                                    elif (
                                        blk.config.split_basis
                                        == SplittingType.phaseFlow
                                    ):
                                        s_vars[v][k].set_value(
                                            value(
                                                m_var[k]
                                                * blk.split_fraction[(t, o) + (k[0],)]
                                            )
                                        )
                                    elif (
                                        blk.config.split_basis
                                        == SplittingType.componentFlow
                                    ):
                                        s_vars[v][k].set_value(
                                            value(
                                                m_var[k]
                                                * blk.split_fraction[(t, o) + (k[1],)]
                                            )
                                        )
                                    else:
                                        raise BurntToast(
                                            "{} encountered unrecognised "
                                            "SplittingType. This should not "
                                            "occur - please send this bug to "
                                            "the IDAES developers.".format(blk.name)
                                        )
                                elif "_phase" in v:
                                    if (
                                        blk.config.split_basis
                                        == SplittingType.phaseComponentFlow
                                    ):
                                        # Need average split fraction
                                        avg_split = value(
                                            sum(
                                                blk.split_fraction[t, o, k, j]
                                                for j in mblock.component_list
                                            )
                                            / len(mblock.component_list)
                                        )
                                        s_vars[v][k].set_value(value(m_var[k] * avg_split))
                                    elif (
                                        blk.config.split_basis
                                        == SplittingType.phaseFlow
                                    ):
                                        s_vars[v][k].set_value(
                                            value(
                                                m_var[k]
                                                * blk.split_fraction[(t, o) + (k,)]
                                            )
                                        )
                                    elif (
                                        blk.config.split_basis
                                        == SplittingType.componentFlow
                                    ):
                                        # Need average split fraction
                                        avg_split = value(
                                            sum(
                                                blk.split_fraction[t, o, j]
                                                for j in mblock.component_list
                                            )
                                            / len(mblock.component_list)
                                        )
                                        s_vars[v][k].set_value(value(m_var[k] * avg_split))
                                    else:
                                        raise BurntToast(
                                            "{} encountered unrecognised "
                                            "SplittingType. This should not "
                                            "occur - please send this bug to "
                                            "the IDAES developers.".format(blk.name)
                                        )
                                elif "_comp" in v:
                                    if (
                                        blk.config.split_basis
                                        == SplittingType.phaseComponentFlow
                                    ):
                                        # Need average split fraction
                                        avg_split = value(
                                            sum(
                                                blk.split_fraction[t, o, p, k]
                                                for p in mblock.phase_list
                                            )
                                            / len(mblock.phase_list)
                                        )
                                        s_vars[v][k].set_value(value(m_var[k] * avg_split))
                                    elif (
                                        blk.config.split_basis
                                        == SplittingType.phaseFlow
                                    ):
                                        # Need average split fraction
                                        avg_split = value(
                                            sum(
                                                blk.split_fraction[t, o, p]
                                                for p in mblock.phase_list
                                            )
                                            / len(mblock.phase_list)
                                        )
                                        s_vars[v][k].set_value(value(m_var[k] * avg_split))
                                    elif (
                                        blk.config.split_basis
                                        == SplittingType.componentFlow
                                    ):
                                        s_vars[v][k].set_value(
                                            value(
                                                m_var[k]
                                                * blk.split_fraction[(t, o) + (k,)]
                                            )
                                        )
                                    else:
                                        raise BurntToast(
                                            "{} encountered unrecognised "
                                            "SplittingType. This should not "
                                            "occur - please send this bug to "
                                            "the IDAES developers.".format(blk.name)
                                        )
                                else:
                                    # Assume unindexed extensive state
                                    # Need average split
                                    if (
                                        blk.config.split_basis
                                        == SplittingType.phaseComponentFlow
                                    ):
                                        # Need average split fraction
                                        avg_split = value(
                                            sum(
                                                blk.split_fraction[t, o, p, j]
                                                for (p, j) in mblock.phase_component_set
                                            )
                                            / len(mblock.phase_component_set)
                                        )
                                    elif (
                                        blk.config.split_basis
                                        == SplittingType.phaseFlow
                                    ):
                                        # Need average split fraction
                                        avg_split = value(
                                            sum(
                                                blk.split_fraction[t, o, p]
                                                for p in mblock.phase_list
                                            )
                                            / len(mblock.phase_list)
                                        )
                                    elif (
                                        blk.config.split_basis
                                        == SplittingType.componentFlow
                                    ):
                                        # Need average split fraction
                                        avg_split = value(
                                            sum(
                                                blk.split_fraction[t, o, j]
                                                for j in mblock.component_list
                                            )
                                            / len(mblock.component_list)
                                        )
                                    else:
                                        raise BurntToast(
                                            "{} encountered unrecognised "
                                            "SplittingType. This should not "
                                            "occur - please send this bug to "
                                            "the IDAES developers.".format(blk.name)
                                        )
                                    s_vars[v][k].set_value(value(m_var[k] * avg_split))
                            else:
                                # Otherwise intensive, equate to mixed stream
                                s_vars[v][k].set_value(m_var[k].value)

            # Call initialization routine for outlet StateBlock
            o_block.initialize(
                outlvl=outlvl,
                optarg=optarg,
                solver=solver,
                hold_state=False,
            )

            # Revert fixed status of variables to what they were before
            for t in blk.flowsheet().time:
                s_vars = o_block[t].define_state_vars()
                for v in s_vars:
                    for k in s_vars[v]:
                        s_vars[v][k].fixed = o_flags[t, v, k]

        if blk.config.mixed_state_block is None:
            with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
                res = opt.solve(blk, tee=slc.tee)

            if not check_optimal_termination(res):
                raise InitializationError(
                    f"{blk.name} failed to initialize successfully. Please "
                    f"check the output logs for more information."
                )

            init_log.info(
                "Initialization Step 2 Complete: {}".format(idaeslog.condition(res))
            )
        else:
            init_log.info("Initialization Complete.")

        if hold_state is True:
            return flags
        else:
            blk.release_state(flags, outlvl=outlvl)
